""" Editor script for x/84 """
# std
import os

# local
from x84.bbs import getsession, getterminal, encode_pipe, echo, getch
from x84.bbs import Lightbar, Selector, ScrollingEditor, showart
from x84.bbs import syncterm_setfont

# This is probably the fourth or more ansi multi-line editor
# I've written for python. I did the least-effort this time.
# There isn't any actual multi-line editor, just this script
# that drives a LineEditor and a Lightbar.

# TODO: use a timer to save the draft -- every 60 seconds or
# so, instead of on every line change.  This should
# significantly improve performance, especially on Raspberry Pi.

WHITESPACE = u' '
SOFTWRAP = u'\n'
HARDWRAP = u'\r\n'
UNDO = list()
UNDOLEVELS = 9

here = os.path.dirname(__file__)

#: preferred fontset for SyncTerm emulator
syncterm_font = 'cp437'


def save_draft(key, ucs):
    """ Persist draft to database and stack changes to UNDO buffer. """
    if key is not None:
        save(key, ucs)

    # pylint: disable=W0602
    #         Using global for 'UNDO' but no assignment is done
    global UNDO
    UNDO.append(ucs)
    if len(UNDO) > UNDOLEVELS:
        del UNDO[0]


def get_contents(lightbar):
    """ Return well-formatted document given the lightbar. """
    return HARDWRAP.join([softwrap_join(_ucs)
                          for _ucs in get_lbcontent(lightbar).split(HARDWRAP)])


def save(key, content):
    """ Persist message to user attrs database. """
    getsession().user[key] = content


def show_help(term):
    """ Returns help text. """
    # clear screen
    echo(term.normal + ('\r\n' * (term.height + 1)) + term.home)

    map(echo, showart(os.path.join(here, 'art', 'po-help.ans')))


def wrap_rstrip(value):
    r""" Remove hardwrap ``u'\r\n'`` and softwrap ``u'\n'`` from value """
    if value[-len(HARDWRAP):] == HARDWRAP:
        value = value[:-len(HARDWRAP)]
    if value[-len(SOFTWRAP):] == SOFTWRAP:
        value = value[:-len(SOFTWRAP)]
    return value


def softwrap_join(value):
    r""" Return whitespace-joined string from value split by softwrap ``'\n'``. """
    return WHITESPACE.join(value.split(SOFTWRAP))


def is_hardwrapped(ucs):
    r""" Returns true if string is hardwrapped with ``'\r\n'``. """
    return ucs[-(len(HARDWRAP)):] == HARDWRAP


def is_softwrapped(ucs):
    r""" Returns true if string is softwrapped with ``'\n'``. """
    return ucs[-(len(SOFTWRAP)):] == SOFTWRAP


def get_lbcontent(lightbar):
    """ Returns ucs string for content of Lightbar instance, ``lightbar``. """
    # a custom 'soft newline' versus 'hard newline' is implemented,
    # '\n' == 'soft', '\r\n' == 'hard'
    lines = list()
    for lno, (_, ucs) in enumerate(lightbar.content):
        # first line always appends as-is, otherwise if the previous line
        # matched a hardwrap, or did not match softwrap, then append as-is.
        # (a simple .endswith() can't wll work with a scheme of '\n' vs.
        # '\r\n')
        if lno == 0 or (
                is_hardwrapped(lines[-1]) or not is_softwrapped(lines[-1])):
            lines.append(ucs)
        else:
            # otherwise the most recently appended line must end with
            # SOFTWRAP, strip that softwrap and re-assign value to a
            # whitespace-joined value by current line value.
            lines[-1] = WHITESPACE.join((lines[-1].rstrip(), ucs.lstrip(),))
    retval = encode_pipe(u''.join(lines))
    return retval


def set_lbcontent(lightbar, ucs):
    """ Sets content for given Unicode string, ``ucs``. """
    # a custom 'soft newline' versus 'hard newline' is implemented,
    # '\n' == 'soft', '\r\n' == 'hard'
    term = getterminal()
    content = dict()
    lno = 0
    lines = ucs.split(HARDWRAP)
    for idx, ucs_line in enumerate(lines):
        if idx == len(lines) - 1 and 0 == len(ucs_line):
            continue
        ucs_joined = WHITESPACE.join(ucs_line.split(SOFTWRAP))
        ucs_wrapped = term.wrap(text=ucs_joined, width=lightbar.visible_width)
        for inner_lno, inner_line in enumerate(ucs_wrapped):
            softwrap = SOFTWRAP if inner_lno != len(ucs_wrapped) - 1 else u''
            content[lno] = u''.join((inner_line, softwrap))
            lno += 1
        if 0 == len(ucs_wrapped):
            content[lno] = HARDWRAP
            lno += 1
        else:
            content[lno - 1] += HARDWRAP
    if 0 == len(content):
        content[0] = HARDWRAP
    lightbar.update(sorted(content.items()))


def yes_no(lightbar, msg, prompt_msg='are you sure? ', attr=None):
    """ Prompt user for yes/no, returns True for yes, False for no. """
    term = getterminal()
    keyset = {
        'yes': (u'y', u'Y'),
        'no': (u'n', u'N'),
    }
    echo(u''.join((
        lightbar.border(),
        lightbar.pos(lightbar.yloc + lightbar.height - 1, lightbar.xpadding),
        msg, u' ', prompt_msg,)))
    sel = Selector(yloc=lightbar.yloc + lightbar.height - 1,
                   xloc=term.width - 25, width=18,
                   left='Yes', right=' No ')
    sel.colors['selected'] = term.reverse_red if attr is None else attr
    sel.keyset['left'].extend(keyset['yes'])
    sel.keyset['right'].extend(keyset['no'])
    echo(sel.refresh())
    while True:
        inp = getch()
        echo(sel.process_keystroke(inp))
        if((sel.selected and sel.selection == sel.left)
                or inp in keyset['yes']):
            # selected 'yes',
            return True
        elif((sel.selected or sel.quit)
                or inp in keyset['no']):
            # selected 'no'
            return False


def get_lightbar(ucs):
    """ Returns lightbar with content of given Unicode string, ``ucs``. """
    term = getterminal()
    width = min(80, max(term.width, 40))
    yloc = 0
    height = term.height - yloc - 1
    xloc = max(0, (term.width / 2) - (width / 2))
    lightbar = Lightbar(height, width, yloc, xloc)
    lightbar.glyphs['left-vert'] = lightbar.glyphs['right-vert'] = u''
    lightbar.colors['highlight'] = term.yellow_reverse
    set_lbcontent(lightbar, ucs)
    return lightbar


def get_lneditor(lightbar):
    """ Returns editor positioned at location of current selection. """
    term = getterminal()
    width = min(80, max(term.width, 40))
    yloc = (lightbar.yloc + lightbar.ypadding + lightbar.position[0] - 1)
    xloc = max(0, (term.width / 2) - (width / 2))
    lneditor = ScrollingEditor(width=width, yloc=yloc, xloc=xloc)
    lneditor.enable_scrolling = True
    lneditor.max_length = 65534
    lneditor.glyphs['bot-horiz'] = u''
    lneditor.glyphs['top-horiz'] = u''
    lneditor.colors['highlight'] = term.red_reverse
    lneditor.colors['border'] = term.bold_red
    # converts u'xxxxxx\r\n' to 'xxxxxx',
    # or 'zzzz\nxxxxxx\n' to u'zzzz xxxxxx',
    lneditor.update(softwrap_join(wrap_rstrip(lightbar.selection[1])))
    return lneditor


def main(save_key=None, continue_draft=False):
    """
    Main Editor procedure.

    When argument ``save_key`` is non-None, the result is saved
    to the user attribute of the same name.  When unset, the
    contents are returned to the caller.

    When argument ``continue_draft`` is non-None, the editor
    continues a previously saved draft, whose contents is its
    value.
    """
    # pylint: disable=R0914,R0912,R0915
    #         Too many local variables
    #         Too many branches
    #         Too many statements
    session, term = getsession(), getterminal()

    # set syncterm font, if any
    if term.kind.startswith('ansi'):
        echo(syncterm_setfont(syncterm_font))

    movement = (term.KEY_UP, term.KEY_DOWN, term.KEY_NPAGE,
                term.KEY_PPAGE, term.KEY_HOME, term.KEY_END,
                u'\r', term.KEY_ENTER)
    keyset = {'edit': (term.KEY_ENTER,),
              'command': (unichr(27), term.KEY_ESCAPE),
              'kill': (u'K',),
              'undo': (u'u', 'U',),
              'goto': (u'G',),
              'insert': (u'I',),
              'insert-before': (u'O',),
              'insert-after': (u'o',),
              'join': (u'J',),
              'rubout': (unichr(8), unichr(127),
                         unichr(23), term.KEY_BACKSPACE,),
              }

    def merge():
        """
        Merges line editor content as a replacement for the currently
        selected lightbar row. Returns True if text inserted caused
        additional rows to be appended, which is meaningful in a window
        refresh context.
        """
        lightbar.content[lightbar.index] = [
            lightbar.selection[0],
            softwrap_join(wrap_rstrip(lneditor.content))
            + HARDWRAP]
        prior_length = len(lightbar.content)
        prior_position = lightbar.position
        set_lbcontent(lightbar, get_lbcontent(lightbar))
        if len(lightbar.content) - prior_length == 0:
            echo(lightbar.refresh_row(prior_position[0]))
            return False
        while len(lightbar.content) - prior_length > 0:
            # hidden move-down for each appended line
            lightbar.move_down()
            prior_length += 1
        return True

    def statusline(lightbar):
        """
        Display status line and command help on ``lightbar`` borders
        """
        lightbar.colors['border'] = term.red if edit else term.yellow
        keyset_cmd = u''
        if not edit:
            keyset_cmd = u''.join((
                term.yellow(u'-( '),
                term.yellow_underline(u'S'), u':', term.bold(u'ave'),
                u' ',
                term.yellow_underline(u'A'), u':', term.bold(u'bort'),
                u' ',
                term.yellow_underline(u'?'), u':', term.bold(u'help'),
                term.yellow(u' )-'),))
            keyset_xpos = max(0, lightbar.width -
                              (term.length(keyset_cmd) + 3))
            keyset_cmd = lightbar.pos(lightbar.yloc + lightbar.height - 1,
                                      keyset_xpos
                                      ) + keyset_cmd
        return u''.join((
            lightbar.border(),
            keyset_cmd,
            lightbar.pos(lightbar.yloc + lightbar.height - 1,
                         lightbar.xpadding),
            u''.join((
                term.red(u'-[ '),
                u'EditiNG liNE ',
                term.reverse_red('%d' % (lightbar.index + 1,)),
                term.red(u' ]-'),)) if edit else u''.join((
                    term.yellow(u'-( '),
                    u'liNE ',
                    term.yellow('%d/%d ' % (
                        lightbar.index + 1,
                        len(lightbar.content),)),
                    '%3d%% ' % (
                        int((float(lightbar.index + 1)
                             / max(1, len(lightbar.content))) * 100)),
                    term.yellow(u' )-'),)),
            lightbar.title(u''.join((
                term.red('-] '),
                term.bold(u'Escape'),
                u':', term.bold_red(u'command mode'),
                term.red(' [-'),)
            ) if edit else u''.join((
                term.yellow('-( '),
                term.bold(u'Enter'),
                u':', term.bold_yellow(u'edit mode'),
                term.yellow(' )-'),))),
            lightbar.fixate(),
        ))

    def redraw_lneditor(lightbar, lneditor):
        """
        Return ucs suitable for refreshing line editor.
        """
        return ''.join((
            term.normal,
            statusline(lightbar),
            lneditor.border(),
            lneditor.refresh()))

    def get_ui(ucs, lightbar=None):
        """
        Returns Lightbar and ScrollingEditor instance.
        """
        lbr = get_lightbar(ucs)
        lbr.position = (lightbar.position
                        if lightbar is not None else (0, 0))
        lne = get_lneditor(lbr)
        return lbr, lne

    def banner():
        """
        Returns string suitable clearing screen
        """
        return u''.join((
            term.move(0, 0),
            term.normal,
            term.clear))

    def redraw(lightbar, lneditor):
        """
        Returns ucs suitable for redrawing Lightbar
        and ScrollingEditor UI elements.
        """
        return u''.join((
            term.normal,
            redraw_lightbar(lightbar),
            redraw_lneditor(lightbar, lneditor) if edit else u'',
        ))

    def redraw_lightbar(lightbar):
        """ Returns ucs suitable for redrawing Lightbar. """
        return u''.join((
            statusline(lightbar),
            lightbar.refresh(),))

    def resize(lightbar):
        """ Resize Lightbar. """
        if edit:
            # always re-merge current line on resize,
            merge()
        lbr, lne = get_ui(get_lbcontent(lightbar), lightbar)
        echo(redraw(lbr, lne))
        return lbr, lne

    if continue_draft:
        ucs = continue_draft
    else:
        ucs = u''
    lightbar, lneditor = get_ui(ucs, None)
    echo(banner())
    dirty = True
    edit = False
    digbuf, num_repeat = u'', -1
    count_repeat = lambda: range(max(num_repeat, 1))
    while True:
        # poll for refresh
        if session.poll_event('refresh'):
            echo(banner())
            lightbar, lneditor = resize(lightbar)
            dirty = True
        if dirty:
            session.activity = 'editing %s' % (save_key,)
            echo(redraw(lightbar, lneditor))
            dirty = False
        # poll for input
        inp = getch(1)

        # buffer keystrokes for repeat
        if (not edit and inp is not None
                and not isinstance(inp, int)
                and inp.isdigit()):
            digbuf += inp
            if len(digbuf) > 10:
                # overflow,
                echo(u'\a')
                digbuf = inp
            try:
                num_repeat = int(digbuf)
            except ValueError:
                try:
                    num_repeat = int(inp)
                except ValueError:
                    pass
            continue
        else:
            digbuf = u''

        # toggle edit mode,
        if inp in keyset['command'] or not edit and inp in keyset['edit']:
            edit = not edit  # toggle
            if not edit:
                # switched to command mode, merge our lines

                echo(term.normal + lneditor.erase_border())

                merge()
                lightbar.colors['highlight'] = term.yellow_reverse
            else:
                # switched to edit mode, save draft,
                # instantiate new line editor
                save_draft(save_key, get_lbcontent(lightbar))
                lneditor = get_lneditor(lightbar)
                lightbar.colors['highlight'] = term.red_reverse
            dirty = True

        # command mode, kill line
        elif not edit and inp in keyset['kill']:
            # when 'killing' a line, make accomidations to clear
            # bottom-most row, otherwise a ghosting effect occurs
            for _ in count_repeat():
                del lightbar.content[lightbar.index]
                set_lbcontent(lightbar, get_lbcontent(lightbar))
            if lightbar.visible_bottom > len(lightbar.content):
                echo(lightbar.refresh_row(lightbar.visible_bottom + 1))
            else:
                dirty = True
            save_draft(save_key, get_lbcontent(lightbar))

        # command mode, insert line
        elif not edit and inp in keyset['insert']:
            for _ in count_repeat():
                lightbar.content.insert(lightbar.index,
                                        (lightbar.index, HARDWRAP,))
                set_lbcontent(lightbar, get_lbcontent(lightbar))
            save_draft(save_key, get_lbcontent(lightbar))
            dirty = True

        # command mode; goto line
        elif not edit and inp in keyset['goto']:
            if num_repeat == -1:
                # 'G' alone goes to end of file,
                num_repeat = len(lightbar.content)
            echo(lightbar.goto((num_repeat or 1) - 1))
            echo(statusline(lightbar))

        # command mode; insert-before (switch to edit mode)
        elif not edit and inp in keyset['insert-before']:
            lightbar.content.insert(lightbar.index,
                                    (lightbar.index, HARDWRAP,))
            set_lbcontent(lightbar, get_lbcontent(lightbar))
            edit = dirty = True
            # switched to edit mode, save draft,
            # instantiate new line editor
            lightbar.colors['highlight'] = term.red_reverse
            lneditor = get_lneditor(lightbar)
            save_draft(save_key, get_lbcontent(lightbar))

        # command mode; insert-after (switch to edit mode)
        elif not edit and inp in keyset['insert-after']:
            lightbar.content.insert(lightbar.index + 1,
                                    (lightbar.index + 1, HARDWRAP,))
            set_lbcontent(lightbar, get_lbcontent(lightbar))
            edit = dirty = True
            # switched to edit mode, save draft,
            # instantiate new line editor
            lightbar.colors['highlight'] = term.red_reverse
            lightbar.move_down()
            lneditor = get_lneditor(lightbar)
            save_draft(save_key, get_lbcontent(lightbar))

        # command mode, undo
        elif not edit and inp in keyset['undo']:
            for _ in count_repeat():
                if len(UNDO):
                    set_lbcontent(lightbar, UNDO.pop())
                    dirty = True
                else:
                    echo(u'\a')
                    break

        # command mode, join line
        elif not edit and inp in keyset['join']:
            for _ in count_repeat():
                if lightbar.index + 1 < len(lightbar.content):
                    idx = lightbar.index
                    lightbar.content[idx] = (idx,
                                             WHITESPACE.join((
                                                 lightbar.content[
                                                     idx][1].rstrip(),
                                                 lightbar.content[idx + 1][1].lstrip(),)))
                    del lightbar.content[idx + 1]
                    prior_length = len(lightbar.content)
                    set_lbcontent(lightbar, get_lbcontent(lightbar))
                    if len(lightbar.content) - prior_length > 0:
                        lightbar.move_down()
                    dirty = True
                else:
                    echo(u'\a')
                    break
            if dirty:
                save_draft(save_key, get_lbcontent(lightbar))

        # command mode, basic cmds & movement
        elif not edit and inp is not None:
            if inp in (u'a', u'A',):
                if yes_no(lightbar, term.yellow(u'- ')
                          + term.bold_red(u'AbORt')
                          + term.yellow(u' -')):
                    return False
                dirty = True
            elif inp in (u's', u'S',):
                if yes_no(lightbar, term.yellow(u'- ')
                          + term.bold_green(u'SAVE')
                          + term.yellow(u' -'), term.reverse_green):
                    # save contents to user attribtue
                    content = get_contents(lightbar)
                    if not save_key:
                        # return entire message body as return value
                        return content
                    save(save_key, content)
                    return True
                dirty = True
            elif inp in (u'?',):
                show_help(term)
                term.inkey()
                # pager = Pager(lightbar.height, lightbar.width,
                #              lightbar.yloc, lightbar.xloc)
                # pager.update(get_help())
                #pager.colors['border'] = term.bold_blue
                # echo(pager.border() + pager.title(u''.join((
                #    term.bold_blue(u'-( '),
                #    term.white_on_blue(u'r'), u':', term.bold(u'eturn'),
                #    u' ',
                #    term.bold_blue(u' )-'),))))
                #pager.keyset['exit'].extend([u'r', u'R'])
                # pager.read()
                # echo(pager.erase_border())
                dirty = True
            else:
                moved = False
                for _ in count_repeat():
                    echo(lightbar.process_keystroke(inp))
                    moved = lightbar.moved or moved
                if moved:
                    echo(statusline(lightbar))

        # edit mode; movement
        elif edit and inp in movement:
            dirty = merge()
            if inp in (u'\r', term.KEY_ENTER,):
                lightbar.content.insert(lightbar.index + 1,
                                        [lightbar.selection[0] + 1, u''])
                inp = term.KEY_DOWN
                dirty = True
            ucs = lightbar.process_keystroke(inp)
            if lightbar.moved:
                # XXX optimize redraws
                echo(term.normal + lneditor.erase_border())
                echo(ucs)
                lneditor = get_lneditor(lightbar)
                save_draft(save_key, get_lbcontent(lightbar))
                echo(lneditor.border() + lneditor.refresh())

        # edit mode -- append character / backspace
        elif edit and inp is not None:
            if (inp in keyset['rubout']
                    and len(lneditor.content) == 0
                    and lightbar.index > 0):
                # erase past margin,
                echo(term.normal + lneditor.erase_border())
                del lightbar.content[lightbar.index]
                lightbar.move_up()
                set_lbcontent(lightbar, get_lbcontent(lightbar))
                lneditor = get_lneditor(lightbar)
                dirty = True
            else:
                # edit mode, add/delete ch
                echo(lneditor.process_keystroke(inp))
                if lneditor.moved:
                    echo(statusline(lightbar))

        if inp is not None and not isinstance(inp, int) and not inp.isdigit():
            # commands were processed, reset num_repeat to 1
            num_repeat = -1



""" logoff script with 'automsg' for x/84. """


def main():
    """ Main procedure. """
    # pylint: disable=R0914,R0912
    #         Too many local variables
    #         Too many branches
    from x84.bbs import DBProxy, getsession, getterminal, echo
    from x84.bbs import ini, LineEditor, timeago, showart
    from x84.bbs import disconnect, getch
    import time
    import os
    session, term = getsession(), getterminal()
    session.activity = 'logging off'
    handle = session.user.handle or 'anonymous'
    max_user = ini.CFG.getint('nua', 'max_user')
    prompt_msg = u'[spnG]: ' if session.user.get('expert', False) else (
        u'%s:AY SOMEthiNG %s:REViOUS %s:EXt %s:Et thE fUCk Off !\b' % (
            term.bold_blue_underline(u's'),
            term.bold_blue_underline(u'p'),
            term.bold_blue_underline(u'n'),
            term.red_underline(u'Escape/g'),))
    prompt_say = u''.join((term.bold_blue(handle),
                           term.bold_blue(u' SAYS WhAt'), term.bold(': '),))
    boards = (('1984.ws', 'x/84 dEfAUlt bOARd', 'Dingo',),
              ('htc.zapto.org', 'Haunting the Chapel', 'Mercyful fate',),
              ('bbs.pharcyde.org', 'Pharcyde BBS', 'Access Denied',),
              ('bloodisland.ph4.se', 'Blood Island/X', 'Xzippo',),
              ('blackflag.acid.org:2627', 'Piranha: Black Flag', 'Caphood',),
              ('oddnetwork.org', '79 columns', 'Haliphax'),
              ('bbs.beardy.se', 'The Swamp', 'Beardy'),
              ('bbs.maze.io', 'rANDOM nOIZE', 'maze'),)
    board_fmt = u'%25s %-30s %-15s\r\n'
    goodbye_msg = u''.join((
        term.move(term.height, 0),
        u'\r\n' * 10,
        u'tRY ANOthER fiNE bOARd', term.bold(u':'), u'\r\n\r\n',
        board_fmt % (
            term.underline('host'.rjust(25)),
            term.underline('board'.ljust(30)),
            term.underline('sysop'.ljust(15)),),
        u'\r\n'.join([board_fmt % (
            term.bold(host.rjust(25)),
            term.reverse(board.center(30)),
            term.bold_underline(sysop),)
            for (host, board, sysop) in boards]),
        u'\r\n\r\n',
        term.bold(
            u'back to the mundane world...'),
        u'\r\n',))
    commit_msg = term.bold_blue(
        u'-- !  thANk YOU fOR YOUR CONtRibUtiON, bROthER  ! --')
    write_msg = term.red_reverse(
        u'bURNiNG tO ROM, PlEASE WAiT ...')
    db_firstrecord = ((time.time() - 1984,
                       u'B. b.', u'bEhAVE YOURSElVES ...'),)
    automsg_len = 40

    def refresh_prompt(msg):
        """ Refresh automsg prompt using string msg. """
        echo(u''.join((u'\r\n\r\n', term.clear_eol, msg)))

    def refresh_automsg(idx):
        """ Refresh automsg database, display automsg of idx, return idx. """
        session.flush_event('automsg')
        autodb = DBProxy('automsg')
        automsgs = sorted(autodb.values()) if len(autodb) else db_firstrecord
        dblen = len(automsgs)
        # bounds check
        if idx < 0:
            idx = dblen - 1
        elif idx > dblen - 1:
            idx = 0
        tm_ago, handle, msg = automsgs[idx]
        asc_ago = u'%s ago' % (timeago(time.time() - tm_ago))
        disp = u''.join(('\r\n\r\n',
                         term.bold(handle.rjust(max_user)),
                         term.bold_blue(u'/'),
                         term.bold_blue(u'%*d' % (len('%d' % (dblen,)), idx,)),
                         term.bold_blue(u':'),
                         term.bold_green(msg.ljust(automsg_len)),
                         term.bold(u'\\'),
                         term.bold_blue(asc_ago),))
        echo(u'\r\n'.join(term.wrap(disp)))
        return idx

    def refresh_all(idx=None):
        """
        refresh screen, database, and return database index
        """
        echo(u''.join((u'\r\n\r\n', term.clear_eol,)))
        for line in showart(
                os.path.join(os.path.dirname(__file__), 'art', 'logoff.ans'), 'cp437'):
            echo(line)
        idx = refresh_automsg(-1 if idx is None else idx)
        refresh_prompt(prompt_msg)
        return idx

    idx = refresh_all()
    while True:
        if session.poll_event('refresh'):
            idx = refresh_all()
        elif session.poll_event('automsg'):
            refresh_automsg(-1)
            echo(u'\a')  # bel
            refresh_prompt(prompt_msg)
        inp = getch(1)
        if inp in (u'g', u'G', term.KEY_EXIT, unichr(27), unichr(3),):
            # http://www.xfree86.org/4.5.0/ctlseqs.html
            # Restore xterm icon and window title from stack.
            echo(unichr(27) + u'[23;0t')
            echo(goodbye_msg)
            getch(1.5)
            disconnect('logoff.')
        elif inp in (u'n', u'N', term.KEY_DOWN, term.KEY_NPAGE,):
            idx = refresh_automsg(idx + 1)
            refresh_prompt(prompt_msg)
        elif inp in (u'p', u'P', term.KEY_UP, term.KEY_PPAGE,):
            idx = refresh_automsg(idx - 1)
            refresh_prompt(prompt_msg)
        elif inp in (u's', u'S'):
            # new prompt: say something !
            refresh_prompt(prompt_say)
            msg = LineEditor(width=automsg_len).read()
            if msg is not None and msg.strip():
                echo(u''.join((u'\r\n\r\n', write_msg,)))
                autodb = DBProxy('automsg')
                autodb.acquire()
                idx = max([int(ixx) for ixx in autodb.keys()] or [-1]) + 1
                autodb[idx] = (time.time(), handle, msg.strip())
                autodb.release()
                session.send_event('global', ('automsg', True,))
                refresh_automsg(idx)
                echo(u''.join((u'\r\n\r\n', commit_msg,)))
                getch(0.5)  # for effect, LoL
            # display prompt
            refresh_prompt(prompt_msg)



""" Common interface utility functions for x/84. """
# std imports
from __future__ import division
import os
import math

# local
from x84.bbs import echo, showart, get_ini
from x84.bbs import getterminal, LineEditor


def decorate_menu_item(menu_item, colors):
    """ Return menu item decorated. """
    key_text = (u'{lb}{inp_key}{rb}'.format(
        lb=colors['lowlight'](u'['),
        rb=colors['lowlight'](u']'),
        inp_key=colors['highlight'](menu_item.inp_key)))

    # set the inp_key within the key_text if matching
    if menu_item.text.startswith(menu_item.inp_key):
        return menu_item.text.replace(menu_item.inp_key, key_text, 1)

    # otherwise prefixed with space
    return (u'{key_text} {menu_text}'.format(
        key_text=key_text, menu_text=menu_item.text))


def render_menu_entries(term, top_margin, menu_items,
                        colors=None, max_cols=3, max_rowsp=2):
    """
    Return all menu items rendered in decorated tabular format.

    :param term: terminal instance returned by :func:`getterminal`.
    :param int top_margin: the top-most row location to begin.
    :param menu_items: any object containing attributes ``inp_key``
                       and ``text``.
    :param dict colors: optional terminal attributes, containing
                        keys of ``highlight`` and ``lowlight``.
    :param int max_cols: maximum number of columns rendered.
    :param int max_row_spacing: maximum vertical row spacing.
    :rtype: str
    """
    # we take measured effects to do this operation much quicker when
    # colored_menu_items is set False to accommodate slower systems
    # such as the raspberry pi.
    if colors is not None:
        measure_width = term.length
    else:
        measure_width = str.__len__
        colors = {}
    colors['highlight'] = colors.get('highlight', lambda txt: txt)
    colors['lowlight'] = colors.get('lowlight', lambda txt: txt)

    # render all menu items, highlighting their action 'key'
    rendered_menuitems = [decorate_menu_item(menu_item, colors)
                          for menu_item in menu_items]

    # create a parallel array of their measurable width
    column_widths = map(measure_width, rendered_menuitems)

    # here, we calculate how many vertical sections of menu entries
    # may be displayed in 80 columns or less -- and forat accordingly
    # so that they are left-adjusted in 1 or more tabular columns, with
    # sufficient row spacing to padd out the full vertical height of the
    # window.
    #
    # It's really just a bunch of math to make centered, tabular columns..
    display_width = min(term.width, 80)
    padding = max(column_widths) + 3
    n_columns = min(max(1, int(math.floor(display_width / padding))), max_cols)
    xpos = max(1, int(math.floor((term.width / 2) - (display_width / 2))))
    xpos += int(math.floor((display_width - ((n_columns * padding))) / 2))
    rows = int(math.ceil(len(rendered_menuitems) / n_columns))
    height = int(math.ceil((term.height - 3) - top_margin))
    row_spacing = min(max(1, min(3, int(math.floor(height / rows)))), max_rowsp)

    column = 1
    output = u''
    for idx, item in enumerate(rendered_menuitems):
        padding_left = term.move_x(xpos) if column == 1 and xpos else u''
        padding_right = ' ' * (padding - column_widths[idx])
        if idx == len(rendered_menuitems) - 1:
            # last item, two newlines
            padding_right = u'\r\n' * 2
        elif column == n_columns:
            # newline(s) on last column only
            padding_right = u'\r\n' * row_spacing
        column = 1 if column == n_columns else column + 1
        output = u''.join((output, padding_left, item, padding_right))
    return output


def waitprompt(term):
    """ Display simple "press enter to continue prompt". """
    echo(u''.join((
        term.normal, '\r\n',
        term.move_x(max(0, (term.width // 2) - 40)),
        term.magenta('('), term.green('..'),
        'press any key to continue', term.green('..'), term.magenta(')')
    )))
    term.inkey()
    return


def display_banner(filepattern, vertical_padding=0, **kwargs):
    """
    Start new screen and show artwork, centered.

    :param str filepattern: file to display
    :param int vertical_padding: number of blank lines to prefix art
    :return: number of lines displayed
    :rtype: int

    Remaining parameters are inherited from :func:`showart`, such
    as ``center`` and ``encoding``.  By default, ``center`` is True.
    """
    # This is unfortunate, we should use 'term' as first argument
    term = getterminal()
    kwargs['center'] = kwargs.get('center', True)

    # move to bottom of screen, reset attribute
    echo(term.move(term.height, 0) + term.normal)

    # create a new, empty screen
    echo(u'\r\n' * (term.height + 1))

    # move to home, insert vertical padding
    echo(term.home + (u'\r\n' * vertical_padding))

    art_generator = showart(filepattern, **kwargs)
    line_no = 0
    for line_no, txt in enumerate(art_generator):
        echo(txt)

    # return line number
    return line_no + vertical_padding


def prompt_pager(content, line_no=0, colors=None, width=None,
                 breaker=u'- ', end_prompt=True, **kwargs):
    """ Display text, using a stop/continuous/next-page prompt.

    :param iterable content: iterable of text contents.
    :param int line_no: line number to offset beginning of pager.
    :param dict colors: optional dictionary containing terminal styling
                        attributes, for keys ``'highlight'`` and
                        ``'lowlight'``.  When unset, yellow and green
                        are used.
    :param int width: When set, text is left-justified-centered by width.
    :param str breaker: repeated decoration for page breaks
    :param bool end_prompt: Whether prompt should be displayed at end.
    :param kwargs: additional arguments to :func:`textwrap.wrap`
    :param bool end_prompt: use 'press enter prompt' at end.
    """
    # This is unfortunate, we should use 'term' as first argument
    term = getterminal()
    colors = colors or {
        'highlight': term.yellow,
        'lowlight': term.green
    }
    pager_prompt = (u'{bl}{s}{br}top, {bl}{c}{br}ontinuous, or '
                    u'{bl}{enter}{br} for next page {br} {bl}\b\b'
                    .format(bl=colors['lowlight'](u'['),
                            br=colors['lowlight'](u']'),
                            s=colors['highlight'](u's'),
                            c=colors['highlight'](u'c'),
                            enter=colors['highlight'](u'return')))

    should_break = lambda line_no, height: line_no % (height - 3) == 0

    def show_breaker():
        if not breaker:
            return u''
        attr = colors['highlight']
        breaker_bar = breaker * (min(80, term.width - 1) // len(breaker))
        echo(attr(term.center(breaker_bar).rstrip()))

    continuous = False

    # we must parse the entire tree, so that we can avoid the needless
    # call to show_breaker() on the final line.
    result = []
    for txt in content:
        if txt.rstrip():
            result.extend(term.wrap(txt, width, **kwargs))
        else:
            result.append(u'\r\n')

    xpos = 0
    if term.width:
        xpos = max(0, int((term.width / 2) - width / 2))
    for line_no, txt in enumerate(result):
        if xpos:
            echo(term.move_x(xpos))
        echo(txt.rstrip() + term.normal + term.clear_eol + u'\r\n')
        if (line_no and line_no != len(result) - 1
                and not continuous
                and should_break(line_no, term.height)):
            show_breaker()
            echo(u'\r\n')
            if xpos:
                echo(term.move_x(xpos))
            echo(pager_prompt)
            while True:
                inp = LineEditor(1, colors=colors).read()
                if inp is None or inp and inp.lower() in u'sqx':
                    # s/q/x/escape: quit
                    echo(u'\r\n')
                    return
                if len(inp) == 1:
                    echo(u'\b')
                if inp.lower() == 'c':
                    # c: enable continuous
                    continuous = True
                    break
                elif inp == u'':
                    # return: next page
                    break
            # remove pager
            echo(term.move_x(0) + term.clear_eol)
            if breaker:
                # and breaker,
                echo(term.move_up() + term.clear_eol)

    if end_prompt:
        show_breaker()
        echo(u'\r\n')
        if term.width > 80:
            echo(term.move_x(max(0, (term.width // 2) - 40)))
        echo(u'Press {enter}.'.format(
            enter=colors['highlight'](u'enter')))
        inp = LineEditor(0, colors=colors).read()


def prompt_input(term, key, sep_ok=u'::', width=None, colors=None):
    """ Prompt for and return input, up to given width and colorscheme. """
    colors = colors or {'highlight': term.yellow}
    sep_ok = colors['highlight'](sep_ok)

    echo(u'{sep} {key:<8}: '.format(sep=sep_ok, key=key))
    return LineEditor(colors=colors, width=width).read() or u''


def coerce_terminal_encoding(term, encoding):
    """ Coerce encoding of terminal to match session by CSI. """
    echo(u'\r\n')
    echo({
        # ESC %G activates UTF-8 with an unspecified implementation
        # level from ISO 2022 in a way that allows to go back to
        # ISO 2022 again.
        'utf8': u'\x1b%G',
        # ESC %@ returns to ISO 2022 in case UTF-8 had been entered.
        # ESC (U Sets character set G0 to codepage 437, such as on
        # Linux vga console.
        'cp437': u'\x1b%@\x1b(U',
    }.get(encoding, u''))
    # remove possible artifacts, at least, %G may print a raw G
    echo(term.move_x(0) + term.clear_eol)


def show_description(term, description, color='white', width=80, **kwargs):
    """
    Display text as given ``color``, left-adjusted ``width``.

    :param str description: description text, may contain terminal attributes,
                            in which case ``color`` should be set to None.
    :param str color: terminal color attribute name, may be None.
    :param int width: left-adjusted width, if this is greater than the current
                      terminal's width, the terminal's width is used instead.
    :param kwargs: all remaining keyword arguments are passed to the built-in
                   :class:`textwrap.TextWrapper`.
    :rtype: int
    :returns: number if lines written
    """
    wide = min(width, term.width)
    xpos = max(0, (term.width // 2) - (wide // 2))

    lines = []
    for line in unicode(description).splitlines():
        if line.strip():
            lines.extend(term.wrap(line, wide, **kwargs))
        else:
            lines.append(u'')

    # output as a single string, reducing latency
    outp = u''.join(
        [getattr(term, color) if color else u''] +
        [u''.join((
            term.move_x(xpos) if xpos else u'',
            txt.rstrip(),
            term.clear_eol,
            u'\r\n')) for txt in lines])
    echo(outp)
    return len(outp.splitlines())


def filesize(filename):
    """ display a file's size in human-readable format """
    size = float(os.stat(filename).st_size)
    for scale in u'BKMGT':
        if size < 1000 or scale == u'T':
            if scale in u'BK':
                # no precision for bytees or kilobytes
                return (u'{size:d}{scale}'
                        .format(size=int(size), scale=scale))
            # 2-decimal precision
            return (u'{size:0.2f}{scale}'
                    .format(size=size, scale=scale))
        size /= 1024


def display_prompt(term, colors):
    """ Return string for displaying a system-wide command prompt. """
    colors['lowlight'] = colors.get('lowlight', lambda txt: txt)
    bbsname = get_ini(section='system', key='bbsname') or 'Unnamed'
    xpos = 0
    if term.width > 30:
        xpos = max(5, int((term.width / 2) - (80 / 2)))
    return (u'{xpos}{user}{at}{bbsname}{colon} '.format(
        xpos=term.move_x(xpos),
        user=term.session.user.handle,
        at=colors['lowlight'](u'@'),
        bbsname=bbsname,
        colon=colors['lowlight'](u'::')))



""" IRC chat script for x/84. """
# std imports
import collections
import warnings
import datetime
import logging
import time
import re

# local
from x84.bbs import ScrollingEditor, getsession, echo, getterminal
from x84.bbs import get_ini, syncterm_setfont, LineEditor

# 3rd party
import irc.client

#: how many lines of scrollback to save for redrawing the interface?
MAX_SCROLLBACK = get_ini('irc', 'max_scrollback', getter='getint') or 200

#: how many characters of input to allow?
MAX_INPUT = get_ini('irc', 'max_input', getter='getint') or 200

#: irc server
SERVER = get_ini('irc', 'server') or 'irc.shaw.ca'

#: irc port
PORT = get_ini('irc', 'port', getter='getint') or 6667

#: irc channel
CHANNEL = get_ini('irc', 'channel') or '#1984'

#: maximum length irc nicks
MAX_NICK = get_ini('irc', 'max_nick', getter='getint') or 9

#: whether irc server requires ssl
ENABLE_SSL = get_ini('irc', 'ssl', getter='getboolean')

#: Whether to display PRIVNOTICE messages (such as motd)
ENABLE_PRIVNOTICE = get_ini('irc', 'enable_privnotice', getter='getboolean')

#: color of line editor
COLOR_INPUTBAR = get_ini('irc', 'color_inputbar') or 'reverse'

#: syncterm font to use
SYNCTERM_FONT = get_ini('irc', 'syncterm_font') or 'topaz'


class IRCChat(object):

    """
    IRC client, based on irc.client.SimpleIRCClient

    This is a class instead of a collection of functions at script-level scope
    because this allows us to easily use the _dispatcher trick... this means
    that we don't have to use add_global_handler() for each and every event we
    want to respond to; just add on_<eventname> as a method and you're done!
    """

    # pylint: disable=R0904

    def __init__(self, term, session):
        """ Class initializer. """
        self.log = logging.getLogger(__name__)
        self.reactor = irc.client.Reactor()
        self.connection = self.reactor.server()
        self.connected = False
        self.reactor.add_global_handler('all_events', self._dispatcher, -10)
        self.term = term
        self.session = session
        self.nick = None

    def _mirc_convert(self, text):
        """ Convert mIRC formatting codes into term sequence equivalents. """
        # pylint: disable=R0201

        def color_repl(match):
            """ Regex function for replacing color codes """

            mirc_colors = [
                'bold_white', 'black', 'blue', 'green', 'bold_red', 'red',
                'magenta', 'yellow', 'bold_yellow', 'bold_green', 'cyan',
                'bold_cyan', 'bold_blue', 'bold_magenta', 'bold_black',
                'white',
            ]
            num_colors = len(mirc_colors)
            bgc = None
            fgc = int(match.group(1))
            boldbg = False
            attr = ''
            if match.group(2):
                bgc = int(match.group(2))
                if bgc >= num_colors:
                    bgc = None
                else:
                    attr = 'on_{0}'.format(mirc_colors[bgc])
                    if 'bold_' in attr:
                        attr = attr.replace('bold_', '')
                        boldbg = True
            if fgc >= num_colors:
                fgc = None
            else:
                attr = '{0}{1}{2}'.format(mirc_colors[fgc],
                                          '_' if attr else '', attr)
            if boldbg:
                attr = 'blink_{0}'.format(attr)
            return getattr(self.term, attr)

        def mode_repl(match):
            """ Regex function for replacing 'modes'. """
            translate = {
                '\x0f': self.term.normal,
                '\x02': self.term.bold,
                '\x1f': self.term.underline,
                '\x15': self.term.underline,
                '\x12': self.term.reverse,
                '\x16': self.term.reverse,
            }
            return u'{0}{1}'.format(translate[match.group(1)], match.group(2))

        text = re.sub(r'\x03(\d{1,2})(?:,(\d{1,2}))?', color_repl, text)
        text = text.replace('\x03', self.term.normal)
        text = re.sub('(\x0f|\x02|\x1f|\x15|\x12|\x16)(.*?)', mode_repl, text)
        text = u''.join([text, self.term.normal])
        return text

    def _dispatcher(self, connection, event):
        """ Dispatch IRC events to their handlers (if they exist) """

        do_nothing = lambda c, e: None
        method = getattr(self, "on_" + event.type, do_nothing)
        method(connection, event)

    def queue(self, message):
        """ Queue up a message (with timestamp) to be displayed """
        now = datetime.datetime.now()
        data = u'{0} {1}'.format(
            self.term.bold_black(u'{0}:{1}'.format(
                '%02d' % now.hour, '%02d' % now.minute)),
            self._mirc_convert(message))
        self.session.buffer_event('irc', (data,))

    def _indicator(self, color=None, character=None):
        """ Construct an indicator to be used in output """
        # pylint: disable=R0201
        color = color or self.term.white
        character = character or u'*'
        return (u'{0}{1}{0}'.format(
            self.term.bold_black(u'-'), color(character)))

    def help(self):
        """ Show /HELP text """

        # pylint: disable=W0141
        commands = map(
            self.term.bold, [
                u'/HELP', u'/ME', u'/TOPIC', u'/NAMES'])
        self.queue(u'{0} Use {1} to quit. Other commands: {2}'.format(
            self._indicator(),
            self.term.bold(u'/QUIT'),
            self.term.white(u', ').join(commands),
        ))

    def connect(self, *args, **kwargs):
        """ Try to connect to the server """

        self.connection.connect(*args, **kwargs)

    def on_disconnect(self, connection, event):
        """ Disconnected; send quit event to end the main loop """
        # pylint: disable=R0201,W0613
        why = filter(None, event.arguments + [event.target])
        self.session.buffer_event('irc-quit', ' '.join(why))

    def on_nicknameinuse(self, connection, event):
        """ Nick is being used; pick another one. """
        # pylint: disable=W0613
        self.session.buffer_event('irc-connected')
        self.connected = True
        echo(u''.join([self.term.normal, u'\r\n',
                       u'Your nickname is in use or illegal. Pick a new one '
                       u'(blank to quit):\r\n']))
        led = LineEditor(width=MAX_NICK)
        newnick = led.read()
        echo(u'\r\n')
        if not newnick:
            self.session.buffer_event('irc-quit', 'canceled')
            return
        connection.nick(newnick)

    def on_erroneusnickname(self, connection, event):
        """ Erroneous nickname; pick another one. """
        return self.on_nicknameinuse(connection, event)

    def on_welcome(self, connection, event):
        """ Connected to the server; join the channel """
        # pylint: disable=W0613
        if not self.connected:
            self.session.buffer_event('irc-connected')
        self.nick = connection.get_nickname()
        self.session.buffer_event('irc-welcome')
        self.help()
        self.queue(u'{0} Server: {1}'.format(
            self._indicator(),
            self.term.bold(self.connection.server)
        ))
        if irc.client.is_channel(CHANNEL):
            connection.join(CHANNEL)

    def on_quit(self, _, event):
        """ Someone has /QUIT IRC """
        self.queue(u'{0} {1} has quit ({2})'.format(
            self._indicator(color=self.term.red),
            self.term.bold(event.source.nick),
            self.term.bold(event.arguments[0])
        ))

    def on_namreply(self, _, event):
        """ Reply received from /NAMES command """
        nicks = list()
        for nick in event.arguments[2].split(' '):
            stripped_nick = (nick[1:]
                             if nick[0] in ('@', '+',)
                             else nick)
            if stripped_nick != self.nick:
                nicks.append(nick)
        self.queue(u'{0} Users: {1}'.format(
            self._indicator(),
            self.term.bold(u', '.join(nicks) if nicks else u'<nobody>')
        ))

    def on_currenttopic(self, _, event):
        """ Reply received from /TOPIC command """
        self.queue(u'{0} Topic: {1}'.format(
            self._indicator(),
            self.term.bold(event.arguments[1])
        ))

    def on_topic(self, _, event):
        """ Someone has changed the channel topic """
        self.queue(u'{0} {1} has changed the topic to {2}'.format(
            self._indicator(color=self.term.bold),
            self.term.bold(event.source.nick),
            self.term.bold(event.arguments[0])
        ))

    def on_privnotice(self, _, event):
        """ Received a private notice. """
        if ENABLE_PRIVNOTICE:
            self.queue('{0} {1}'.format(event.target, event.arguments[0]))

    def format_pubmsg(self, source, msg, mode=None):
        """ Helper to format public messages. """
        # pylint: disable=R0201
        color = self.term.bold if mode is None else mode
        nick = color(u'<{0}>'.format(source))
        return u'{0} {1}'.format(nick, msg)

    def on_pubmsg(self, _, event):
        """ Public message has been received. """
        color = None
        if self.nick in event.arguments[0]:
            color = self.term.bold_white_on_magenta
        self.queue(
            self.format_pubmsg(event.source.nick, event.arguments[0], color))

    def on_join(self, connection, event):
        """ Someone has joined the channel """
        # if it's the current user that has joined, display it differently
        if event.source.nick == connection.get_nickname():
            self.queue(u'{0} Channel: {1}'.format(
                self._indicator(),
                self.term.bold(CHANNEL)
            ))
        else:
            self.queue(u'{0} {1} has joined the channel'.format(
                self._indicator(color=self.term.green),
                self.term.bold(event.source.nick)
            ))

    def on_kick(self, connection, event):
        """ Someone has been kicked from the channel """
        self.queue(u'{0} {1} has kicked {2} ({3})'.format(
            self._indicator(color=self.term.red),
            self.term.bold(event.source.nick),
            self.term.bold(event.arguments[0]),
            self.term.bold(event.arguments[1])
        ))

        # if the current user was kicked, rejoin the channel
        if event.arguments[0] == self.nick:
            connection.join(CHANNEL)

    def on_part(self, _, event):
        """ Someone has /PART-ed the channel """
        self.queue(u'{0} {1} has left the channel'.format(
            self._indicator(color=self.term.red),
            self.term.bold(event.source.nick)
        ))

    def format_action(self, source, action, color=None):
        """ Helper for formatting action messages """
        color = self.term.bold_white if color is None else color
        return u'{0} {1} {2}'.format(
            self._indicator(character=u'+', color=self.term.bold_black),
            color(source), action)

    def on_action(self, _, event):
        """ Someone has performed an action in the channel """
        color = None
        if self.nick in event.arguments[0]:
            color = self.term.bold_white_on_magenta
        self.queue(
            self.format_action(event.source.nick, event.arguments[0], color))

    def on_nick(self, _, event):
        """ Someone has changed their nickname """
        self.queue(u'{0} {1} has changed their nickname to {2}'.format(
            self._indicator(),
            self.term.bold(event.source.nick),
            self.term.bold(event.target)
        ))

    def on_mode(self, _, event):
        """ Someone has changed modes on the channel or a user """
        self.queue(u'{0} {1} sets mode {2}'.format(
            self._indicator(color=self.term.bold),
            self.term.bold(event.source.nick),
            self.term.bold(u' '.join(event.arguments))
        ))

    def on_pubnotice(self, _, event):
        """ Public notice has been received """
        self.queue(u'{0} NOTICE ({1}) {2}'.format(
            self._indicator(),
            self.term.bold(event.source.nick),
            self.term.bold(u' '.join(event.arguments))
        ))

    def on_error(self, _, event):
        """ Some error has been received """
        err_desc = (u'error (type={event.type}, source={event.source}, '
                    'target={event.target}, arguments={event.arguments}'
                    .format(event=event))
        self.log.error('error: {0}'.format(err_desc))
        self.queue(u'{0} ERROR: {1} '.format(
            self._indicator(color=self.term.bold_red), err_desc))


def clean_up(term):
    """ Clean up after ourselves """
    echo(term.normal)
    echo(term.clear)


def establish_connection(term, session):
    """
    Establish a connection to the IRC server.

    Returns IRCChat instance or False depending on whether it was successful.
    """
    scrollback = collections.deque(maxlen=MAX_SCROLLBACK)
    kwargs = dict()
    if ENABLE_SSL:
        from ssl import wrap_socket
        from irc.connection import Factory
        kwargs['connect_factory'] = Factory(wrapper=wrap_socket)

    echo(u'Connecting to {server}:{port} for channel {chan}.\r\n\r\n'
         'press [{key_q}] to abort ... '.format(
             server=SERVER, port=PORT, chan=CHANNEL, key_q=term.bold(u'q')))

    client = IRCChat(term, session)
    irc_handle = session.user.handle.replace(' ', '_')
    try:
        # pylint: disable=W0142
        client.connect(SERVER, PORT, irc_handle, **kwargs)
    except irc.client.ServerConnectionError:
        echo(term.bold_red(u'Connection error!'))
        term.inkey(3)
        raise EOFError()

    while True:
        client.reactor.process_once()
        event, data = session.read_events(
            ('irc', 'irc-quit', 'irc-connected', 'input'), timeout=0.02)
        if event == 'irc':
            # show on-connect motd data if any received
            echo(u'\r\n{0}'.format(data[0]))
            scrollback.append(data[0])
        elif event == 'irc-connected':
            break
        elif event == 'input':
            session.buffer_input(data, pushback=True)
            inp = term.inkey(0)
            while inp is not None:
                if inp.lower() == u'q' or inp.code == term.KEY_ESCAPE:
                    echo(u'Canceled!')
                    raise EOFError()
                inp = term.inkey(0)
        elif event == 'irc-quit':
            echo(term.bold_red(u'\r\nConnection failed: {0}!'.format(data)))
            term.inkey(3)
            raise EOFError()

    while True:
        client.reactor.process_once()
        event, data = session.read_events(
            ('irc', 'irc-welcome', 'irc-quit', 'input'), timeout=0.2)
        if event == 'irc':
            # show on-connect motd data if any received
            echo(u'\r\n{0}'.format(data[0]))
            scrollback.append(data[0])
        if event == 'irc-quit':
            echo(term.bold_red(u'\r\nConnection lost: {0}!'.format(data)))
            term.inkey(3)
            raise EOFError()
        elif event == 'input':
            session.buffer_input(data, pushback=True)
            inp = term.inkey(0)
            while inp is not None:
                if inp.lower() == u'q' or inp.code == term.KEY_ESCAPE:
                    echo(u'Canceled!')
                    raise EOFError()
                inp = term.inkey(0)
        elif event == 'irc-welcome':
            return client, scrollback


def refresh_event(term, scrollback, editor):
    """ Screen resized, adjust layout """

    editor.width = term.width + 1
    editor.xloc = -1
    editor.yloc = term.height - 1

    # re-wrap chat log
    numlines = term.height - 2
    sofar = 0
    output = list()
    for line in reversed(scrollback):
        wrapped = term.wrap(line, term.width - 1)
        sofar += len(wrapped)
        output = wrapped + output
        if sofar >= numlines:
            break
    output = output[numlines * -1:]
    echo(u''.join([
        term.normal, term.home, term.clear_eos,
        term.move(editor.yloc + 1, 0),
        u'\r\n'.join(output),
        u'\r\n', editor.refresh()]))


def irc_event(term, data, scrollback, editor):
    """ IRC output has been received """

    data = data[0]
    # add it to our scrollback and trim
    scrollback.append(data)

    # blank out the line with the input bar, wrap output, redraw input bar
    echo(u''.join([
        term.normal,
        term.move(editor.yloc + 1, 0),
        term.clear_eol,
        term.move_x(0),
        u'\r\n'.join(term.wrap(data, term.width - 1,
                               break_long_words=True)),
        term.move(editor.yloc + 1, 0),
        u'\r\n',
        term.move(editor.yloc, 0),
        editor.refresh(),
    ]))


def input_event(term, client, editor):
    """
    React to input event, processing /commands.

    Returns True, unless the caller should exit, then False.
    """
    retval = True
    inp = term.inkey(0)
    while inp:
        if inp.is_sequence:
            echo(editor.process_keystroke(inp.code))
        else:
            echo(editor.process_keystroke(inp))

        if not editor.carriage_returned:
            retval = True
        else:
            # process return key
            line = editor.content.rstrip()

            # clear editor
            editor.update(u'')
            echo(editor.refresh())

            lowered = line.lower()

            # fix mIRC codes that are sent improperly
            line = (line
                    .replace('\x15', '\x1f')   # underline
                    .replace('\x12', '\x16'))  # reverse

            # parse input for potential commands
            if lowered == u'/quit':
                client.connection.quit(
                    u'x/84 bbs https://github.com/jquast/x84')
                retval = False
            elif lowered == u'/topic':
                client.connection.topic(CHANNEL)
            elif lowered in (u'/names', u'/who',):
                client.connection.names(CHANNEL)
            elif lowered.startswith(u'/me ') and len(lowered) > 4:
                client.connection.action(CHANNEL, line[4:])
                client.queue(client.format_action(client.nick, line[4:],
                                                  term.bold_blue))
            elif lowered == u'/help':
                client.help()
            elif line.startswith(u'/'):
                # some other .. unsupported command
                retval = True
            else:
                # no command was received; post pubmsg, instead
                client.connection.privmsg(CHANNEL, line)
                client.queue(client.format_pubmsg(
                    client.nick, line, term.bold_blue))
        inp = term.inkey(0)
    return retval


def main():
    """ x/84 script launch point """

    term, session = getterminal(), getsession()
    session.activity = u'Chatting on IRC'
    session.flush_event('irc')

    # move to bottom of screen, reset attribute
    echo(term.pos(term.height) + term.normal)

    # create a new, empty screen
    echo(u'\r\n' * (term.height + 1))
    echo(term.home + term.clear)

    # set font
    if SYNCTERM_FONT and term.kind.startswith('ansi'):
        echo(syncterm_setfont(SYNCTERM_FONT))

    try:
        client, scrollback = establish_connection(term, session)
    except EOFError:
        # connection failed,
        return clean_up(term)

    # ignore "not in view" warning for AnsiWindow
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        editor = ScrollingEditor(
            width=term.width + 1,
            xloc=-1,
            yloc=term.height,
            colors={'highlight': getattr(term, COLOR_INPUTBAR)},
            max_length=MAX_INPUT
        )

    refresh_event(term, scrollback, editor)

    while True:
        client.reactor.process_once()
        event, data = session.read_events(
            ('irc-quit', 'irc', 'input', 'refresh'), timeout=0.1)
        if event == 'refresh':
            refresh_event(term, scrollback, editor)
            continue
        elif event == 'irc':
            irc_event(term, data, scrollback, editor)
        elif event == 'input':
            session.buffer_input(data, pushback=True)
            if not input_event(term, client, editor):
                break
        elif event == 'irc-quit':
            time.sleep(0.5)
            break
    client.connection.disconnect()
    clean_up(term)



"""
Amiga "Topaz" codec for x/84.

There is no provided unicode translation map, but we try to approximate the
Topaz font families as close as possible.
"""

from __future__ import absolute_import

import codecs
import encodings
import encodings.aliases

# Codec APIs


class Codec(codecs.Codec):

    def decode(self, input, errors='strict'):
        return codecs.charmap_decode(input, errors, decoding_table)

    def encode(self, input, errors='strict'):
        raise NotImplementedError()


class IncrementalEncoder(codecs.IncrementalEncoder):

    def encode(self, input, final=False):
        raise NotImplementedError()


class IncrementalDecoder(codecs.IncrementalDecoder):

    def decode(self, input, final=False):
        return codecs.charmap_decode(input, self.errors, decoding_table)[0]


class StreamReader(Codec, codecs.StreamReader):
    pass


class StreamWriter(Codec, codecs.StreamWriter):
    pass


# encodings module API

def getaliases():
    return (
        'microknight',
        'mosoul',
        'p0tnoodle',
        'topaz',
        'topaz1',
        'topaz1plus',
        'topaz2',
        'topaz2plus',
        'topazplus',
    )


def getregentry():
    return codecs.CodecInfo(
        name='topaz',
        encode=Codec().encode,
        decode=Codec().decode,
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=StreamWriter,
    )


encodings._cache['topaz'] = getregentry()


# Decoding Table

decoding_map = codecs.make_identity_dict(range(256))
decoding_map.update({
    0x002d: 0x2500,     # BOX DRAWINGS LIGHT HORIZONTAL
    0x002f: 0x2571,     # BOX DRAWINGS LIGHT DIAGONAL UPPER RIGHT TO LOWER LEFT
    0x0058: 0x2573,     # BOX DRAWINGS LIGHT DIAGONAL CROSS
    0x005c: 0x2572,     # BOX DRAWINGS LIGHT DIAGONAL UPPER LEFT TO LOWER RIGHT
    0x005f: 0x2581,     # LOWER ONE EIGHTH BLOCK
    0x007c: 0x2502,     # BOX DRAWINGS LIGHT VERTICAL
    0x007f: 0x259e,     # QUADRANT UPPER RIGHT AND LOWER LEFT
    0x0080: 0x2b1c,     # WHITE LARGE SQUARE
    0x0081: 0x2b1c,     # WHITE LARGE SQUARE
    0x0082: 0x2b1c,     # WHITE LARGE SQUARE
    0x0083: 0x2b1c,     # WHITE LARGE SQUARE
    0x0084: 0x2b1c,     # WHITE LARGE SQUARE
    0x0085: 0x2b1c,     # WHITE LARGE SQUARE
    0x0086: 0x2b1c,     # WHITE LARGE SQUARE
    0x0087: 0x2b1c,     # WHITE LARGE SQUARE
    0x0088: 0x2b1c,     # WHITE LARGE SQUARE
    0x0089: 0x2b1c,     # WHITE LARGE SQUARE
    0x008a: 0x2b1c,     # WHITE LARGE SQUARE
    0x008b: 0x2b1c,     # WHITE LARGE SQUARE
    0x008c: 0x2b1c,     # WHITE LARGE SQUARE
    0x008d: 0x2b1c,     # WHITE LARGE SQUARE
    0x008e: 0x2b1c,     # WHITE LARGE SQUARE
    0x008f: 0x2b1c,     # WHITE LARGE SQUARE
    0x0090: 0x2b1c,     # WHITE LARGE SQUARE
    0x0091: 0x2b1c,     # WHITE LARGE SQUARE
    0x0092: 0x2b1c,     # WHITE LARGE SQUARE
    0x0093: 0x2b1c,     # WHITE LARGE SQUARE
    0x0094: 0x2b1c,     # WHITE LARGE SQUARE
    0x0095: 0x2b1c,     # WHITE LARGE SQUARE
    0x0096: 0x2b1c,     # WHITE LARGE SQUARE
    0x0097: 0x2b1c,     # WHITE LARGE SQUARE
    0x0098: 0x2b1c,     # WHITE LARGE SQUARE
    0x0099: 0x2b1c,     # WHITE LARGE SQUARE
    0x009a: 0x2b1c,     # WHITE LARGE SQUARE
    0x009b: 0x2b1c,     # WHITE LARGE SQUARE
    0x009c: 0x2b1c,     # WHITE LARGE SQUARE
    0x009d: 0x2b1c,     # WHITE LARGE SQUARE
    0x009e: 0x2b1c,     # WHITE LARGE SQUARE
    0x009f: 0x2b1c,     # WHITE LARGE SQUARE
    0x00af: 0x2594,     # UPPER ONE EIGHTH BLOCK
})

decoding_table = (
    u'\x00'     # 0x0000 -> NULL
    u'\x01'     # 0x0001 -> START OF HEADING
    u'\x02'     # 0x0002 -> START OF TEXT
    u'\x03'     # 0x0003 -> END OF TEXT
    u'\x04'     # 0x0004 -> END OF TRANSMISSION
    u'\x05'     # 0x0005 -> ENQUIRY
    u'\x06'     # 0x0006 -> ACKNOWLEDGE
    u'\x07'     # 0x0007 -> BELL
    u'\x08'     # 0x0008 -> BACKSPACE
    u'\t'       # 0x0009 -> HORIZONTAL TABULATION
    u'\n'       # 0x000a -> LINE FEED
    u'\x0b'     # 0x000b -> VERTICAL TABULATION
    u'\x0c'     # 0x000c -> FORM FEED
    u'\r'       # 0x000d -> CARRIAGE RETURN
    u'\x0e'     # 0x000e -> SHIFT OUT
    u'\x0f'     # 0x000f -> SHIFT IN
    u'\x10'     # 0x0010 -> DATA LINK ESCAPE
    u'\x11'     # 0x0011 -> DEVICE CONTROL ONE
    u'\x12'     # 0x0012 -> DEVICE CONTROL TWO
    u'\x13'     # 0x0013 -> DEVICE CONTROL THREE
    u'\x14'     # 0x0014 -> DEVICE CONTROL FOUR
    u'\x15'     # 0x0015 -> NEGATIVE ACKNOWLEDGE
    u'\x16'     # 0x0016 -> SYNCHRONOUS IDLE
    u'\x17'     # 0x0017 -> END OF TRANSMISSION BLOCK
    u'\x18'     # 0x0018 -> CANCEL
    u'\x19'     # 0x0019 -> END OF MEDIUM
    u'\x1a'     # 0x001a -> SUBSTITUTE
    u'\x1b'     # 0x001b -> ESCAPE
    u'\x1c'     # 0x001c -> FILE SEPARATOR
    u'\x1d'     # 0x001d -> GROUP SEPARATOR
    u'\x1e'     # 0x001e -> RECORD SEPARATOR
    u'\x1f'     # 0x001f -> UNIT SEPARATOR
    u' '        # 0x0020 -> SPACE
    u'!'        # 0x0021 -> EXCLAMATION MARK
    u'"'        # 0x0022 -> QUOTATION MARK
    u'#'        # 0x0023 -> NUMBER SIGN
    u'$'        # 0x0024 -> DOLLAR SIGN
    u'%'        # 0x0025 -> PERCENT SIGN
    u'&'        # 0x0026 -> AMPERSAND
    u"'"        # 0x0027 -> APOSTROPHE
    u'('        # 0x0028 -> LEFT PARENTHESIS
    u')'        # 0x0029 -> RIGHT PARENTHESIS
    u'*'        # 0x002a -> ASTERISK
    u'+'        # 0x002b -> PLUS SIGN
    u','        # 0x002c -> COMMA
    u'\u2500'   # 0x002d -> BOX DRAWINGS LIGHT HORIZONTAL
    u'.'        # 0x002e -> FULL STOP
    # 0x002f -> BOX DRAWINGS LIGHT DIAGONAL UPPER RIGHT TO LOWER LEFT
    u'\u2571'
    u'0'        # 0x0030 -> DIGIT ZERO
    u'1'        # 0x0031 -> DIGIT ONE
    u'2'        # 0x0032 -> DIGIT TWO
    u'3'        # 0x0033 -> DIGIT THREE
    u'4'        # 0x0034 -> DIGIT FOUR
    u'5'        # 0x0035 -> DIGIT FIVE
    u'6'        # 0x0036 -> DIGIT SIX
    u'7'        # 0x0037 -> DIGIT SEVEN
    u'8'        # 0x0038 -> DIGIT EIGHT
    u'9'        # 0x0039 -> DIGIT NINE
    u':'        # 0x003a -> COLON
    u';'        # 0x003b -> SEMICOLON
    u'<'        # 0x003c -> LESS-THAN SIGN
    u'='        # 0x003d -> EQUALS SIGN
    u'>'        # 0x003e -> GREATER-THAN SIGN
    u'?'        # 0x003f -> QUESTION MARK
    u'@'        # 0x0040 -> COMMERCIAL AT
    u'A'        # 0x0041 -> LATIN CAPITAL LETTER A
    u'B'        # 0x0042 -> LATIN CAPITAL LETTER B
    u'C'        # 0x0043 -> LATIN CAPITAL LETTER C
    u'D'        # 0x0044 -> LATIN CAPITAL LETTER D
    u'E'        # 0x0045 -> LATIN CAPITAL LETTER E
    u'F'        # 0x0046 -> LATIN CAPITAL LETTER F
    u'G'        # 0x0047 -> LATIN CAPITAL LETTER G
    u'H'        # 0x0048 -> LATIN CAPITAL LETTER H
    u'I'        # 0x0049 -> LATIN CAPITAL LETTER I
    u'J'        # 0x004a -> LATIN CAPITAL LETTER J
    u'K'        # 0x004b -> LATIN CAPITAL LETTER K
    u'L'        # 0x004c -> LATIN CAPITAL LETTER L
    u'M'        # 0x004d -> LATIN CAPITAL LETTER M
    u'N'        # 0x004e -> LATIN CAPITAL LETTER N
    u'O'        # 0x004f -> LATIN CAPITAL LETTER O
    u'P'        # 0x0050 -> LATIN CAPITAL LETTER P
    u'Q'        # 0x0051 -> LATIN CAPITAL LETTER Q
    u'R'        # 0x0052 -> LATIN CAPITAL LETTER R
    u'S'        # 0x0053 -> LATIN CAPITAL LETTER S
    u'T'        # 0x0054 -> LATIN CAPITAL LETTER T
    u'U'        # 0x0055 -> LATIN CAPITAL LETTER U
    u'V'        # 0x0056 -> LATIN CAPITAL LETTER V
    u'W'        # 0x0057 -> LATIN CAPITAL LETTER W
    u'\u2573'   # 0x0058 -> BOX DRAWINGS LIGHT DIAGONAL CROSS
    u'Y'        # 0x0059 -> LATIN CAPITAL LETTER Y
    u'Z'        # 0x005a -> LATIN CAPITAL LETTER Z
    u'['        # 0x005b -> LEFT SQUARE BRACKET
    # 0x005c -> BOX DRAWINGS LIGHT DIAGONAL UPPER LEFT TO LOWER RIGHT
    u'\u2572'
    u']'        # 0x005d -> RIGHT SQUARE BRACKET
    u'^'        # 0x005e -> CIRCUMFLEX ACCENT
    u'\u2581'   # 0x005f -> LOWER ONE EIGHTH BLOCK
    u'`'        # 0x0060 -> GRAVE ACCENT
    u'a'        # 0x0061 -> LATIN SMALL LETTER A
    u'b'        # 0x0062 -> LATIN SMALL LETTER B
    u'c'        # 0x0063 -> LATIN SMALL LETTER C
    u'd'        # 0x0064 -> LATIN SMALL LETTER D
    u'e'        # 0x0065 -> LATIN SMALL LETTER E
    u'f'        # 0x0066 -> LATIN SMALL LETTER F
    u'g'        # 0x0067 -> LATIN SMALL LETTER G
    u'h'        # 0x0068 -> LATIN SMALL LETTER H
    u'i'        # 0x0069 -> LATIN SMALL LETTER I
    u'j'        # 0x006a -> LATIN SMALL LETTER J
    u'k'        # 0x006b -> LATIN SMALL LETTER K
    u'l'        # 0x006c -> LATIN SMALL LETTER L
    u'm'        # 0x006d -> LATIN SMALL LETTER M
    u'n'        # 0x006e -> LATIN SMALL LETTER N
    u'o'        # 0x006f -> LATIN SMALL LETTER O
    u'p'        # 0x0070 -> LATIN SMALL LETTER P
    u'q'        # 0x0071 -> LATIN SMALL LETTER Q
    u'r'        # 0x0072 -> LATIN SMALL LETTER R
    u's'        # 0x0073 -> LATIN SMALL LETTER S
    u't'        # 0x0074 -> LATIN SMALL LETTER T
    u'u'        # 0x0075 -> LATIN SMALL LETTER U
    u'v'        # 0x0076 -> LATIN SMALL LETTER V
    u'w'        # 0x0077 -> LATIN SMALL LETTER W
    u'x'        # 0x0078 -> LATIN SMALL LETTER X
    u'y'        # 0x0079 -> LATIN SMALL LETTER Y
    u'z'        # 0x007a -> LATIN SMALL LETTER Z
    u'{'        # 0x007b -> LEFT CURLY BRACKET
    u'\u2502'   # 0x007c -> BOX DRAWINGS LIGHT VERTICAL
    u'}'        # 0x007d -> RIGHT CURLY BRACKET
    u'~'        # 0x007e -> TILDE
    u'\u259e'   # 0x007f -> QUADRANT UPPER RIGHT AND LOWER LEFT
    u'\u2b1c'   # 0x0080 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0081 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0082 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0083 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0084 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0085 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0086 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0087 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0088 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0089 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x008a -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x008b -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x008c -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x008d -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x008e -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x008f -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0090 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0091 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0092 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0093 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0094 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0095 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0096 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0097 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0098 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x0099 -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x009a -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x009b -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x009c -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x009d -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x009e -> WHITE LARGE SQUARE
    u'\u2b1c'   # 0x009f -> WHITE LARGE SQUARE
    u'\xa0'     # 0x00a0 -> NO-BREAK SPACE
    u'\xa1'     # 0x00a1 -> INVERTED EXCLAMATION MARK
    u'\xa2'     # 0x00a2 -> CENT SIGN
    u'\xa3'     # 0x00a3 -> POUND SIGN
    u'\xa4'     # 0x00a4 -> CURRENCY SIGN
    u'\xa5'     # 0x00a5 -> YEN SIGN
    u'\xa6'     # 0x00a6 -> BROKEN BAR
    u'\xa7'     # 0x00a7 -> SECTION SIGN
    u'\xa8'     # 0x00a8 -> DIAERESIS
    u'\xa9'     # 0x00a9 -> COPYRIGHT SIGN
    u'\xaa'     # 0x00aa -> FEMININE ORDINAL INDICATOR
    u'\xab'     # 0x00ab -> LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    u'\xac'     # 0x00ac -> NOT SIGN
    u'\xad'     # 0x00ad -> SOFT HYPHEN
    u'\xae'     # 0x00ae -> REGISTERED SIGN
    u'\xaf'     # 0x00af -> MACRON
    u'\xb0'     # 0x00b0 -> DEGREE SIGN
    u'\xb1'     # 0x00b1 -> PLUS-MINUS SIGN
    u'\xb2'     # 0x00b2 -> SUPERSCRIPT TWO
    u'\xb3'     # 0x00b3 -> SUPERSCRIPT THREE
    u'\xb4'     # 0x00b4 -> ACUTE ACCENT
    u'\xb5'     # 0x00b5 -> MICRO SIGN
    u'\xb6'     # 0x00b6 -> PILCROW SIGN
    u'\xb7'     # 0x00b7 -> MIDDLE DOT
    u'\xb8'     # 0x00b8 -> CEDILLA
    u'\xb9'     # 0x00b9 -> SUPERSCRIPT ONE
    u'\xba'     # 0x00ba -> MASCULINE ORDINAL INDICATOR
    u'\xbb'     # 0x00bb -> RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    u'\xbc'     # 0x00bc -> VULGAR FRACTION ONE QUARTER
    u'\xbd'     # 0x00bd -> VULGAR FRACTION ONE HALF
    u'\xbe'     # 0x00be -> VULGAR FRACTION THREE QUARTERS
    u'\xbf'     # 0x00bf -> INVERTED QUESTION MARK
    u'\xc0'     # 0x00c0 -> LATIN CAPITAL LETTER A WITH GRAVE
    u'\xc1'     # 0x00c1 -> LATIN CAPITAL LETTER A WITH ACUTE
    u'\xc2'     # 0x00c2 -> LATIN CAPITAL LETTER A WITH CIRCUMFLEX
    u'\xc3'     # 0x00c3 -> LATIN CAPITAL LETTER A WITH TILDE
    u'\xc4'     # 0x00c4 -> LATIN CAPITAL LETTER A WITH DIAERESIS
    u'\xc5'     # 0x00c5 -> LATIN CAPITAL LETTER A WITH RING ABOVE
    u'\xc6'     # 0x00c6 -> LATIN CAPITAL LETTER AE
    u'\xc7'     # 0x00c7 -> LATIN CAPITAL LETTER C WITH CEDILLA
    u'\xc8'     # 0x00c8 -> LATIN CAPITAL LETTER E WITH GRAVE
    u'\xc9'     # 0x00c9 -> LATIN CAPITAL LETTER E WITH ACUTE
    u'\xca'     # 0x00ca -> LATIN CAPITAL LETTER E WITH CIRCUMFLEX
    u'\xcb'     # 0x00cb -> LATIN CAPITAL LETTER E WITH DIAERESIS
    u'\xcc'     # 0x00cc -> LATIN CAPITAL LETTER I WITH GRAVE
    u'\xcd'     # 0x00cd -> LATIN CAPITAL LETTER I WITH ACUTE
    u'\xce'     # 0x00ce -> LATIN CAPITAL LETTER I WITH CIRCUMFLEX
    u'\xcf'     # 0x00cf -> LATIN CAPITAL LETTER I WITH DIAERESIS
    u'\xd0'     # 0x00d0 -> LATIN CAPITAL LETTER ETH
    u'\xd1'     # 0x00d1 -> LATIN CAPITAL LETTER N WITH TILDE
    u'\xd2'     # 0x00d2 -> LATIN CAPITAL LETTER O WITH GRAVE
    u'\xd3'     # 0x00d3 -> LATIN CAPITAL LETTER O WITH ACUTE
    u'\xd4'     # 0x00d4 -> LATIN CAPITAL LETTER O WITH CIRCUMFLEX
    u'\xd5'     # 0x00d5 -> LATIN CAPITAL LETTER O WITH TILDE
    u'\xd6'     # 0x00d6 -> LATIN CAPITAL LETTER O WITH DIAERESIS
    u'\xd7'     # 0x00d7 -> MULTIPLICATION SIGN
    u'\xd8'     # 0x00d8 -> LATIN CAPITAL LETTER O WITH STROKE
    u'\xd9'     # 0x00d9 -> LATIN CAPITAL LETTER U WITH GRAVE
    u'\xda'     # 0x00da -> LATIN CAPITAL LETTER U WITH ACUTE
    u'\xdb'     # 0x00db -> LATIN CAPITAL LETTER U WITH CIRCUMFLEX
    u'\xdc'     # 0x00dc -> LATIN CAPITAL LETTER U WITH DIAERESIS
    u'\xdd'     # 0x00dd -> LATIN CAPITAL LETTER Y WITH ACUTE
    u'\xde'     # 0x00de -> LATIN CAPITAL LETTER THORN
    u'\xdf'     # 0x00df -> LATIN SMALL LETTER SHARP S
    u'\xe0'     # 0x00e0 -> LATIN SMALL LETTER A WITH GRAVE
    u'\xe1'     # 0x00e1 -> LATIN SMALL LETTER A WITH ACUTE
    u'\xe2'     # 0x00e2 -> LATIN SMALL LETTER A WITH CIRCUMFLEX
    u'\xe3'     # 0x00e3 -> LATIN SMALL LETTER A WITH TILDE
    u'\xe4'     # 0x00e4 -> LATIN SMALL LETTER A WITH DIAERESIS
    u'\xe5'     # 0x00e5 -> LATIN SMALL LETTER A WITH RING ABOVE
    u'\xe6'     # 0x00e6 -> LATIN SMALL LETTER AE
    u'\xe7'     # 0x00e7 -> LATIN SMALL LETTER C WITH CEDILLA
    u'\xe8'     # 0x00e8 -> LATIN SMALL LETTER E WITH GRAVE
    u'\xe9'     # 0x00e9 -> LATIN SMALL LETTER E WITH ACUTE
    u'\xea'     # 0x00ea -> LATIN SMALL LETTER E WITH CIRCUMFLEX
    u'\xeb'     # 0x00eb -> LATIN SMALL LETTER E WITH DIAERESIS
    u'\xec'     # 0x00ec -> LATIN SMALL LETTER I WITH GRAVE
    u'\xed'     # 0x00ed -> LATIN SMALL LETTER I WITH ACUTE
    u'\xee'     # 0x00ee -> LATIN SMALL LETTER I WITH CIRCUMFLEX
    u'\xef'     # 0x00ef -> LATIN SMALL LETTER I WITH DIAERESIS
    u'\xf0'     # 0x00f0 -> LATIN SMALL LETTER ETH
    u'\xf1'     # 0x00f1 -> LATIN SMALL LETTER N WITH TILDE
    u'\xf2'     # 0x00f2 -> LATIN SMALL LETTER O WITH GRAVE
    u'\xf3'     # 0x00f3 -> LATIN SMALL LETTER O WITH ACUTE
    u'\xf4'     # 0x00f4 -> LATIN SMALL LETTER O WITH CIRCUMFLEX
    u'\xf5'     # 0x00f5 -> LATIN SMALL LETTER O WITH TILDE
    u'\xf6'     # 0x00f6 -> LATIN SMALL LETTER O WITH DIAERESIS
    u'\xf7'     # 0x00f7 -> DIVISION SIGN
    u'\xf8'     # 0x00f8 -> LATIN SMALL LETTER O WITH STROKE
    u'\xf9'     # 0x00f9 -> LATIN SMALL LETTER U WITH GRAVE
    u'\xfa'     # 0x00fa -> LATIN SMALL LETTER U WITH ACUTE
    u'\xfb'     # 0x00fb -> LATIN SMALL LETTER U WITH CIRCUMFLEX
    u'\xfc'     # 0x00fc -> LATIN SMALL LETTER U WITH DIAERESIS
    u'\xfd'     # 0x00fd -> LATIN SMALL LETTER Y WITH ACUTE
    u'\xfe'     # 0x00fe -> LATIN SMALL LETTER THORN
    u'\xff'     # 0x00ff -> LATIN SMALL LETTER Y WITH DIAERESIS
)



"""
Post-login screen for x/84.

When handle is None or u'', an in-memory account 'anonymous' is created
and assigned to the session.
"""
# std
import functools
import logging
import time
import glob
import os

# local
from x84.bbs import getterminal, showart, echo, get_ini
from x84.bbs import getsession, get_user, User, LineEditor
from x84.bbs import goto, gosub, DBProxy, syncterm_setfont
from x84.default.common import (
    coerce_terminal_encoding,
    show_description,
    waitprompt,
)

# 3rd
from sauce import SAUCE

here = os.path.dirname(__file__)
log = logging.getLogger(__name__)

#: glob of art files
art_files = glob.glob(os.path.join(here, 'art', 'top', '*'))

#: encoding used for artwork
art_encoding = 'cp437_art'

#: font used for syncterm
syncterm_font = 'cp437'

#: time to pause while displaying
art_speed = 0.04

#: which sauce records to display
sauce_records = set(['author', 'title', 'group', 'date', 'filename'])

#: ssh port configured, this is displayed to the user, if you're using NAT port
#: forwarding or something like that, you'll want to set the configuration
#: value of section [ssh] for key 'advertise_port'.
ssh_port = (get_ini(section='ssh', key='advertise_port') or
            get_ini(section='ssh', key='port'))

#: maximum number of sauce columns for wide displays
max_sauce_columns = 2

#: iterator returns each line of art with given settings
idisplay_art = functools.partial(showart, encoding=art_encoding,
                                 auto_mode=False, center=True,
                                 poll_cancel=art_speed)


def _show_opt(term, keys):
    """ Display characters ``key`` highlighted as keystroke """
    return u''.join((term.bold_black(u'['),
                     term.bold_red_underline(keys),
                     term.bold_black(u']')))


def get_art_detail(art_file):
    """
    Return list of tuple (key, value) of ``art_file`` details.

    at least key and value of filename is produced, and additional
    details when ``art_file`` contains sauce records.
    """
    parsed = SAUCE(art_file)
    value = ()
    if parsed.record:
        # parse all 'sauce_records' except 'filename',
        value = tuple((attr, getattr(parsed, attr).strip())
                      for attr in sauce_records ^ set(['filename'])
                      if getattr(parsed, attr).strip())
    # inject 'filename' in return value
    return value + (('filename', os.path.basename(art_file)),)


def display_sauce(term, art_file):

    # total number of columns
    sauce_columns = min(len(sauce_records), max_sauce_columns)

    # 'value' column size
    val_adjust = 20

    # 'key' column size
    key_adjust = max(len(val) for val in sauce_records)

    # will the current tabulation fit?
    sauce_width = lambda: (
        ((key_adjust + val_adjust) * sauce_columns)
        + (len(u':' * sauce_columns))
        + (len(u' ') * sauce_columns))

    # resize until it does !
    while sauce_columns > 1 and sauce_width() >= term.width:
        sauce_columns -= 1

    # construct formatted tabular sauce data
    sauce_disp = [u'{key}{colon}{space}{value}'
                  .format(key=term.bold_black(key.rjust(key_adjust)),
                          colon=term.bold_black_underline(u':'),
                          space=u' ',
                          value=term.red(value.ljust(val_adjust)))
                  for key, value in get_art_detail(art_file)]

    # justify last row for alignment
    if sauce_disp:
        while len(sauce_disp) % sauce_columns != 0:
            sauce_disp.append(u'{key}{colon}{space}{value}'
                              .format(key=u' ' * key_adjust,
                                      colon=u' ',
                                      space=u' ',
                                      value=u' ' * val_adjust))

    # display sauce / filename data
    for idx in range(0, len(sauce_disp), sauce_columns):
        echo(term.center(
            u''.join(sauce_disp[idx:idx + sauce_columns])
        ).rstrip() + u'\r\n')


def display_prompt(term):
    """ Display prompt of user choices. """
    # show prompt
    echo(u'\r\n\r\n')
    echo(term.center(u'{0} previous - {1} change encoding - next {2}'
                     .format(_show_opt(term, u'<'),
                             _show_opt(term, u'!'),
                             _show_opt(term, u'>'))
                     ).rstrip())
    echo(term.move_up() + term.move_up() + term.move_x(0))
    echo(term.center(u'quick login {0} ?\b\b'
                     .format(_show_opt(term, u'yn'))
                     ).rstrip())


def display_intro(term, index):
    """ Display art, '!' encoding prompt, and quick login [yn] ? """

    # clear screen
    echo(term.normal)
    echo(term.move_x(0) + term.clear_eol + '\r\n\r\n' + term.clear_eol)
    echo(term.normal + ('\r\n' * (term.height + 1)) + term.home)

    # show art
    art_file = art_files[index % len(art_files)]
    line = u''
    for line in idisplay_art(art_file):
        echo(line)
    if line.strip():
        # write newline only if last line was artful
        echo(u'\r\n')
    display_sauce(term, art_file)
    echo(u'\r\n')


def get_user_record(handle):
    """
    Find and return User class instance by given ``handle``.

    If handle is ``anonymous``, Create and return a new User object.
    """
    if handle == u'anonymous':
        log.debug('anonymous login')
        return User(u'anonymous')

    log.debug('login by {0!r}'.format(handle))
    return get_user(handle)


def login(session, user):
    """
    Assign ``user`` to ``session`` and return time of last call.

    performs various saves and lookups of call records
    """
    session.user = user

    # assign timeout preference
    timeout = session.user.get('timeout', None)
    if timeout is not None:
        session.send_event('set-timeout', timeout)

    # update call records
    user.calls += 1
    user.lastcall = time.time()

    # store "from" (session id, telnet-<host>:<port>, fe.)
    user['last_from'] = session.sid

    # save user record
    if user.handle != u'anonymous':
        user.save()

    # update 'lastcalls' database
    lc_db = DBProxy('lastcalls')
    with lc_db:
        previous_call, _, _ = lc_db.get(user.handle, (0, 0, 0,))
        lc_db[user.handle] = (user.lastcall, user.calls, user.location)

    return previous_call


def do_intro_art(term, session):
    """
    Display random art file, prompt for quick login.

    Bonus: allow chosing other artfiles with '<' and '>'.
    """
    # set syncterm font, if any
    if syncterm_font and term.kind.startswith('ansi'):
        echo(syncterm_setfont(syncterm_font))

    index = int(time.time()) % len(art_files)
    dirty = True
    echo(u'\r\n')
    while True:
        session.activity = 'top'
        if session.poll_event('refresh') or dirty:
            display_intro(term, index)
            display_prompt(term)
            dirty = False
        dirty = True
        inp = LineEditor(1, colors={'highlight': term.normal}).read()
        if inp is None or inp.lower() == u'y':
            # escape/yes: quick login
            return True
        elif inp.lower() == u'n':
            break

        if len(inp) == 1:
            echo(u'\b')
        if inp == u'!':
            echo(u'\r\n' * 3)
            gosub('charset')
            dirty = True
        elif inp == u'<':
            index -= 1
        elif inp == u'>':
            index += 1
        else:
            dirty = False


def describe_ssh_availability(term, session):
    from x84.bbs.ini import CFG
    if session.kind == 'ssh':
        # what a good citizen!
        return

    if not (CFG.has_section('ssh') and
            not CFG.has_option('ssh', 'enabled')
            or CFG.getboolean('ssh', 'enabled')):
        # ssh not enabled
        return

    about_key = (u"You may even use an ssh key, which you can configure from "
                 u"your user profile, " if not session.user.get('pubkey')
                 else u'')
    big_msg = term.bold_blue("Big Brother is Watching You")
    description = (
        u"\r\n\r\n"
        u"    {term.red}You are using {session.kind}, but ssh is available "
        u"on port {ssh_port} of this server.  If you want a secure connection "
        u"with shorter latency, we recommend instead to use ssh!  {about_key}"
        u"Remember: {big_msg}!\r\n\r\n".format(
            term=term, session=session, ssh_port=ssh_port,
            about_key=about_key, big_msg=big_msg))

    show_description(term, description, color=None)
    waitprompt(term)


def main(handle=None):
    """ Main procedure. """
    # pylint: disable=R0914,R0912,R0915
    #         Too many local variables
    #         Too many branches
    #         Too many statements
    session, term = getsession(), getterminal()
    session.activity = 'top'

    # attempt to coerce encoding of terminal to match session.
    coerce_terminal_encoding(term, session.encoding)

    # fetch user record and register, or "log" call
    login(session, get_user_record(handle))

    # display art and prompt for quick login
    quick = do_intro_art(term, session)

    echo(term.move_down() * 3)

    if not quick:
        describe_ssh_availability(term, session)

        gosub('msgarea', quick=True)

        # only display news if the account has not
        # yet read the news since last update.
        gosub('news', quick=True)

        # display last 10 callers
        gosub('lc')

        # one-liners
        gosub('ol')

    goto('main')



"""
Default matrix (login) script for x/84.

This script is the default session entry point for all connections.

Or simply put, the login program. It is configured in default.ini file,
under section 'matrix'. Alternative matrices may be considered by their
connection type, using script_{telnet,ssh}

In legacy era, a matrix script might be something to fool folk not in the
know, meant to divert agents from underground boards, require a passcode,
or even swapping the modem into a strange stop/bit/parity configuration,
callback mechanisms, etc..

Read all about it in old e-zines.
"""
# std
import logging
import random
import time
import os

from x84.bbs import getterminal, get_ini, goto, gosub
from x84.bbs import echo, showart, syncterm_setfont, LineEditor
from x84.bbs import find_user, get_user, User
from x84.engine import __url__

log = logging.getLogger()
here = os.path.dirname(__file__)

#: name of bbs
system_bbsname = get_ini(
    section='system', key='bbsname'
) or 'Unnamed'

#: whether new user accounts are accepted
new_allowed = get_ini(
    section='nua', key='allow_apply', getter='getboolean'
) or False

#: which login names trigger new user account script
new_usernames = get_ini(
    section='matrix', key='newcmds', split=True
) or ['new']

#: which script to execute to apply for new account
new_script = get_ini(
    section='nua', key='script'
) or 'nua'

#: which script to execute on successful login
top_script = get_ini(
    section='matrix', key='topscript'
) or 'top'

#: whether anonymous login is allowed
anonymous_allowed = get_ini(
    section='matrix', key='enable_anonymous', getter='getboolean'
) or False

#: which login names that trigger anonymous login.
anonymous_names = get_ini(
    section='matrix', key='anoncmds', split=True
) or ['anonymous']

#: whether password resets are allowed
reset_allowed = get_ini(
    section='matrix', key='enable_pwreset', getter='getboolean'
) or False

reset_script = get_ini(
    section='matrix', key='reset_script'
) or 'pwreset'

bye_usernames = get_ini(
    section='matrix', key='byecmds', split=True
) or ['bye', 'logoff', 'exit', 'quit']

#: maximum length of user handles
username_max_length = get_ini(
    section='nua', key='max_user', getter='getint'
) or 10

#: maximum length of password
password_max_length = get_ini(
    section='nua', key='max_pass', getter='getint'
) or 15

#: random maximum time to artificially sleep for unknown user.
unknown_sleep = get_ini(
    section='matrix', key='unknown_sleep', getter='getfloat'
) or 2.0

#: maximum failed logins before disconect
login_max_attempts = get_ini(
    section='matrix', key='max_login_attemps', getter='getint'
) or 5

#: on-connect fontset for SyncTerm emulator
syncterm_font = get_ini(
    section='matrix', key='syncterm_font'
) or 'topaz'

#: on-connect banner
art_file = get_ini(
    section='matrix', key='art_file'
) or os.path.join(here, 'art', 'matrix.ans')

#: encoding on banner
art_encoding = get_ini(
    section='matrix', key='art_encoding'
) or 'cp437'

#: primary color (highlight)
color_primary = get_ini(
    section='matrix', key='color_primary'
) or 'red'

#: secondary color (lowlight)
color_secondary = get_ini(
    section='matrix', key='color_secondary'
) or 'green'

#: password hidden character
hidden_char = get_ini(
    section='nua', key='hidden_char'
) or u'\u00f7'


def display_banner(term):
    """ Display on-connect banner and set a few sequences. """

    # reset existing SGR attributes
    echo(term.normal)

    # set syncterm font, if any
    if syncterm_font and term.kind.startswith('ansi'):
        echo(syncterm_setfont(syncterm_font))

    # http://www.termsys.demon.co.uk/vtansi.htm
    # disable line-wrapping (SyncTerm does not honor, careful!)
    echo(u'\x1b[7l')

    if term.kind.startswith('xterm'):
        # http://www.xfree86.org/4.5.0/ctlseqs.html
        # Save xterm icon and window title on stack.
        echo(u'\x1b[22;0t')

    # move to beginning of line and clear, in case syncterm_setfont
    # has been mis-interpreted, as it follows CSI with space, which
    # causes most terminal emulators to receive literally after CSI.
    echo(term.move_x(0) + term.clear_eol)

    # display name of bbs and url to sourcecode.
    highlight = getattr(term, color_primary)
    sep = getattr(term, color_secondary)(u'::')

    echo(u'{sep} Connected to {name}.\r\n'.format(
        sep=sep, name=highlight(system_bbsname)))
    echo(u'{sep} See {url} for source code.\r\n'.format(
        sep=sep, url=highlight(__url__)))

    # display on-connect banner (`art_file`)
    map(echo, showart(art_file, encoding=art_encoding, center=True))

    # display various ini-configured login username aliases.
    if new_allowed:
        echo(u"   Login as '{0}' to create an account."
             .format(highlight(new_usernames[0])))
    if anonymous_allowed:
        echo(u"\r\n   Login as '{0}' is allowed."
             .format(highlight(anonymous_names[0])))
    if reset_allowed:
        echo(u"\r\n   Forgot password? Login as '{0}'."
             .format(highlight('reset')))


def authenticate_user(handle, password):
    """ Return True if the given handle and password are correct. """

    # artificial delay -- this ensures people cannot guess
    # for user accounts, where existing ones would delay a
    # long while, but unknown users are quickly denied.
    artificial_delay = max(1.0, random.randrange(0, unknown_sleep * 100) / 100)

    matching_handle = find_user(handle)
    if matching_handle is None:
        log.debug('Failed login for {handle}: no such user.'
                  .format(handle=handle))
        time.sleep(artificial_delay)
        return False

    elif not password.strip():
        log.debug('Failed login for {handle}: password not provided.'
                  .format(handle=handle))
        time.sleep(artificial_delay)
        return False

    user = get_user(matching_handle)
    if user.auth(password):
        # success !
        log.debug('Login succeeded for {handle}.'
                  .format(handle=handle))
        return user

    log.debug('Failed login for {handle}: wrong password.'
              .format(handle=handle))
    return False


def do_login(term):
    sep_ok = getattr(term, color_secondary)(u'::')
    sep_bad = getattr(term, color_primary)(u'::')
    colors = {'highlight': getattr(term, color_primary)}
    for _ in range(login_max_attempts):
        echo(u'\r\n\r\n{sep} Login: '.format(sep=sep_ok))
        handle = LineEditor(username_max_length, colors=colors
                            ).read() or u''

        if handle.strip() == u'':
            continue

        # user says goodbye
        if handle.lower() in bye_usernames:
            return

        # user applies for new account
        if new_allowed and handle.lower() in new_usernames:
            gosub(new_script)
            display_banner(term)
            continue

        # user wants to reset password
        if reset_allowed and handle.lower() == 'reset':
            gosub(reset_script)
            display_banner(term)
            continue

        # user wants to login anonymously
        if anonymous_allowed and handle.lower() in anonymous_names:
            user = User('anonymous')
        else:
            # authenticate password
            echo(u'\r\n\r\n{sep} Password: '.format(sep=sep_ok))
            password = LineEditor(password_max_length,
                                  colors=colors,
                                  hidden=hidden_char
                                  ).read() or u''

            user = authenticate_user(handle, password)
            if not user:
                echo(u'\r\n\r\n{sep} Login failed.'.format(sep=sep_bad))
                continue

        goto(top_script, handle=user.handle)

    echo(u'\r\n\r\n{sep} Too many authentication attempts.\r\n'
         .format(sep=sep_bad))


def main(anonymous=False, new=False):
    """
    Script entry point.

    This is the default login matrix for the bbs system.

    It takes no arguments or keyword arguments, because it assumes
    the user should now be authenticated, such as occurs for example
    on telnet.
    """
    term = getterminal()

    display_banner(term)

    if anonymous:
        # user rlogin'd in as anonymous@
        goto(top_script, 'anonymous')
    elif new:
        # user rlogin'd in as new@
        goto(new_script)

    # do_login will goto/gosub various scripts, if it returns, then
    # either the user entered 'bye', or had too many failed attempts.
    do_login(term)

    log.debug('Disconnecting.')

    # it is necessary to provide sufficient time to send any pending
    # output across the transport before disconnecting.
    term.inkey(1.5)



""" Ansi Windowing package for x/84. """

GLYPHSETS = {
    'ascii': {
        'top-left': u'+',
        'bot-left': u'+',
        'top-right': u'+',
        'bot-right': u'+',
        'left-vert': u'|',
        'right-vert': u'|',
        'top-horiz': u'-',
        'bot-horiz': u'-', },
    'thin': {
        'top-left': chr(218).decode('cp437'),
        'bot-left': chr(192).decode('cp437'),
        'top-right': chr(191).decode('cp437'),
        'bot-right': chr(217).decode('cp437'),
        'left-vert': chr(179).decode('cp437'),
        'right-vert': chr(179).decode('cp437'),
        'top-horiz': chr(196).decode('cp437'),
        'bot-horiz': chr(196).decode('cp437'),
    },
}


class AnsiWindow(object):

    """
    Provides position-relative drawing routines within a region.

    The AnsiWindow base class provides position-relative window drawing
    routines to terminal interfaces, such as pager windows, editors, and
    lightbar lists, as well as some drawing niceties such as borders,
    text alignment
    """

    # pylint: disable=R0902,R0913
    #        Too many instance attributes
    #        Too many arguments

    def __init__(self, height, width, yloc, xloc, colors=None, glyphs=None):
        """
        Class initializer for base windowing class.

        :param int width: width of window.
        :param int height: height of window.
        :param int yloc: y-location of window.
        :param int xloc: x-location of window.
        :param dict colors: color theme.
        :param dict glyphs: bordering window character glyphs.
        """
        from x84.bbs.session import getterminal
        self._term = getterminal()

        self.height = height
        self.width = width
        self.yloc = yloc
        self.xloc = xloc
        self.init_theme(colors, glyphs)

        self._xpadding = 1
        self._ypadding = 1
        self._alignment = 'left'
        self._moved = False

    def init_theme(self, colors=None, glyphs=None):
        """
        Set glyphs and colors appropriate for "theming".

        This is called by the class initializer.
        """
        # set defaults,
        self.colors = {
            'normal': self._term.normal,
            'border': (self._term.number_of_colors and
                       self._term.cyan or self._term.normal),
        }
        self.glyphs = GLYPHSETS['thin'].copy()
        # allow user override
        if colors is not None:
            self.colors.update(colors)
        if glyphs is not None:
            self.glyphs.update(glyphs)

    @property
    def xpadding(self):
        """ Horizontal padding of window border. """
        return self._xpadding

    @xpadding.setter
    def xpadding(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._xpadding = value

    @property
    def ypadding(self):
        """ Vertical padding of window border. """
        return self._ypadding

    @ypadding.setter
    def ypadding(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._ypadding = value

    @property
    def alignment(self):
        """ Horizontal justification of text content for method ``align``. """
        return self._alignment

    @alignment.setter
    def alignment(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        assert value in ('left', 'right', 'center')
        self._alignment = value

    def align(self, text, width=None):
        """
        Return ``text`` aligned to ``width`` using ``self.alignment``.

        When None (default), the visible width of this window is used.
        """
        width = width if width is not None else (self.visible_width)
        return (self._term.rjust(text, width) if self.alignment == 'right' else
                self._term.ljust(text, width) if self.alignment == 'left' else
                self._term.center(text, width)
                )

    @property
    def visible_height(self):
        """ Visible height of window after accounting for padding.  """
        return self.height - (self.ypadding * 2)

    @property
    def visible_width(self):
        """ Visible width of window after accounting for padding. """
        return self.width - (self.xpadding * 2)

    def resize(self, height=None, width=None, yloc=None, xloc=None):
        """ Adjust window dimensions by given parameter. """
        if height is not None:
            self.height = height
        if width is not None:
            self.width = width
        if yloc is not None:
            self.yloc = yloc
        if xloc is not None:
            self.xloc = xloc

    def isinview(self):
        """ Whether this window is in bounds of terminal dimensions. """
        return (self.xloc >= 0
                and self.xloc + self.width <= self._term.width
                and self.yloc >= 0
                and self.yloc + self.height <= self._term.height)

    def willfit(self, win):
        """ Whether target window, ``win`` is within this windows bounds. """
        return (win.yloc >= self.yloc
                and win.yloc + win.height <= self.yloc + self.height
                and win.xloc >= self.xloc
                and win.xloc + win.w <= self.xloc + self.width)

    # deprecated alias
    iswithin = willfit

    def pos(self, yloc=None, xloc=None):
        """ Return sequence to move cursor to window-relative position.  """
        return self._term.move((yloc and yloc or 0) + self.yloc,
                               (xloc and xloc or 0) + self.xloc)

    def title(self, ansi_text):
        """ Return sequence for displaying text on top border of window. """
        xloc = self.width / 2 - min(self._term.length(ansi_text) / 2,
                                    self.width / 2)
        return self.pos(0, max(0, xloc)) + ansi_text

    def footer(self, text):
        """ Return sequence for displaying text on bottom border of window. """
        xloc = self.width / 2 - min(self._term.length(text) / 2,
                                    self.width / 2)
        return self.pos(max(0, self.height - 1), max(0, xloc)) + text

    def border(self):
        """ Return sequence suitable for drawing window border. """
        # pylint: disable=R0912
        #        Too many branches (17/12)
        topright = self.glyphs.get('top-right', u'*')
        thoriz = self.glyphs.get('top-horiz', u'-') * (max(0, self.width - 2))
        topleft = self.glyphs.get('top-left', u'/')
        leftvert = self.glyphs.get('left-vert', u'|')
        rightvert = self.glyphs.get('right-vert', u'|')
        botleft = self.glyphs.get('bot-left', u'\\')
        bhoriz = self.glyphs.get('bot-horiz', u'-') * (max(0, self.width - 2))
        botright = self.glyphs.get('bot-right', u'/')
        rstr = u''
        for row in range(0, self.height):
            for col in range(0, self.width):
                if col == 0 or col == self.width - 1:
                    rstr += (self.pos(row, col) + topleft
                             if row == 0 and col == 0 else
                             self.pos(row, col) + botleft
                             if row == self.height - 1 and col == 0 else
                             self.pos(row, col) + topright
                             if row == 0 else
                             self.pos(row, col) + botright
                             if row == self.height - 1 else
                             self.pos(row, col) + leftvert
                             if col == 0 else
                             self.pos(row, col) + rightvert
                             if col == self.width - 1 else
                             u'')
                elif row == 0:
                    # top row (column 1)
                    if thoriz == u'':
                        if topright != u'':
                            # prepare for top-right, (horiz skipped)
                            rstr += self.pos(row, max(0, self.width - 1))
                    else:
                        rstr += thoriz
                    rstr += topright
                    break
                elif row == self.height - 1:
                    # bottom row (column 1)
                    if bhoriz == u'':
                        if botright != u'':
                            # prepare for bot-right, (horiz skipped)
                            rstr += self.pos(row, max(0, self.width - 1))
                    else:
                        # horizontal line
                        rstr += bhoriz
                    # bot-right
                    rstr += botright
                    break
        return (self.colors.get('border', u'') + rstr +
                self.colors.get('normal', u''))

    def erase_border(self):
        """ Return sequence suitable for erasing only the window border. """
        save = self.glyphs.copy()
        for glyph in ('top-left', 'top-horiz', 'top-right',
                      'left-vert', 'right-vert',
                      'bot-left', 'bot-horiz', 'bot-right'):
            if 0 != len(self.glyphs.get(glyph, u'')):
                self.glyphs[glyph] = self.glyphs.get('erase', u' ')
        ucs = self.border()
        # pylint: disable=W0201
        #         Attribute 'glyphs' defined outside __init__
        self.glyphs = save
        return ucs

    def erase(self):
        """ Return sequence suitable for erasing full window (with border). """
        return u''.join([self.pos(y, 0)
                         + (self.glyphs.get('erase', u' ') * self.width)
                         for y in range(self.height)
                         ])

    def clear(self):
        """ Return sequence suitable for erasing contents window. """
        return u''.join([self.pos(self.ypadding + yloc, self.xpadding)
                         + (self.glyphs.get('erase', u' ')
                            * self.visible_width)
                         for yloc in range(self.visible_height)
                         ])

    @property
    def moved(self):
        """ Whether movement has occurred (bool). """
        return self._moved

    @moved.setter
    def moved(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._moved = value



""" Left/Right lightbar choice selector for x/84. """
from x84.bbs.ansiwin import AnsiWindow

VI_KEYSET = {
    'refresh': [unichr(12)],
    'toggle': [u' '],
    'left': [u'h'],
    'right': [u'l'],
    'enter': [u'\r'],
    'exit': [u'q', u'Q', unichr(27), ],
}


class Selector(AnsiWindow):

    """ A two-state horizontal lightbar interface. """

    # pylint: disable=R0902,R0913,R0904
    #        Too many instance attributes (8/7)
    #        Too many arguments (6/5)
    #        Too many public methods (25/20)

    def __init__(self, yloc, xloc, width, left, right, **kwargs):
        """
        Class initializer.

        Initialize a selector of width, y x, and left/right values.

        :param int width: width of window.
        :param int yloc: y-location of selector.
        :param int xloc: x-location of selector.
        :param dict colors: color theme, only key value of ``selected``
                            and ``unselected`` is used.
        :param dict keyset: command keys, global ``VI_KEYSET`` is
                            used by default, augmented by application
                            keys such as home, end, pgup, etc.
        :param str left: text string of left-side selection.
        :param str right: text string of right-side selection.
        """
        self._left = self._selection = left
        self._right = right
        self._moved = False
        self._quit = False
        self._selected = False

        self.init_keystrokes(keyset=kwargs.pop('keyset', VI_KEYSET.copy()))

        AnsiWindow.__init__(self, height=1, width=width,
                            yloc=yloc, xloc=xloc, **kwargs)

    def init_theme(self, colors=None, glyphs=None):
        from x84.bbs.session import getterminal
        term = getterminal()
        colors = colors or {
            'selected': term.reverse_yellow,
            'unselected': term.bold_black,
        }
        AnsiWindow.init_theme(self, colors=colors, glyphs=glyphs)

    def init_keystrokes(self, keyset):
        """ Sets keyboard keys for various editing keystrokes. """
        from x84.bbs.session import getterminal
        self.keyset = keyset
        term = getterminal()
        self.keyset['refresh'].append(term.KEY_REFRESH)
        self.keyset['left'].append(term.KEY_LEFT)
        self.keyset['right'].append(term.KEY_RIGHT)
        self.keyset['enter'].append(term.KEY_ENTER)
        self.keyset['exit'].append(term.KEY_ESCAPE)

    def process_keystroke(self, keystroke):
        """
        Process the keystroke and return string to refresh.

        :param blessed.keyboard.Keystroke keystroke: input from ``inkey()``.
        :rtype: str
        :returns: string sequence suitable for refresh.
        """
        self._moved = False
        keystroke = hasattr(keystroke, 'code') and keystroke.code or keystroke
        if keystroke in self.keyset['refresh']:
            return self.refresh()
        elif keystroke in self.keyset['left']:
            return self.move_left()
        elif keystroke in self.keyset['right']:
            return self.move_right()
        elif keystroke in self.keyset['toggle']:
            return self.toggle()
        elif keystroke in self.keyset['exit']:
            self._quit = True
        elif keystroke in self.keyset['enter']:
            self._selected = True
        return u''

    def read(self):
        """ Reads input until the ENTER or ESCAPE key is pressed (Blocking). """
        from x84.bbs import getch
        from x84.bbs.output import echo
        self._selected = False
        self._quit = False
        echo(self.refresh())
        while not (self.selected or self.quit):
            echo(self.process_keystroke(getch()) or u'')
        if self.quit:
            return None
        return self.selection

    @property
    def selected(self):
        """ Whether the carriage return character has been handled. """
        return self._selected

    @selected.setter
    def selected(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        # this setter should only be used to reset to 'False' for recycling
        assert isinstance(value, bool)
        self._selected = value

    @property
    def selection(self):
        """ Current selection. """
        return self._selection

    @selection.setter
    def selection(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        assert value in (self._left, self._right)
        if self._selection != value:
            self._moved = True
            self._selection = value

    @property
    def left(self):
        """ Left-side value. """
        return self._left

    @left.setter
    def left(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._left = value

    @property
    def right(self):
        """ Right-side value. """
        return self._right

    @right.setter
    def right(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._right = value

    def refresh(self):
        """ Return string sequence suitable for refresh. """
        import math
        import x84.bbs.session
        term = x84.bbs.session.getterminal()
        attr_l = (self.colors.get('selected', u'')
                  if self.selection == self.left
                  else self.colors.get('unselected'))
        attr_r = (self.colors.get('selected', u'')
                  if self.selection == self.right
                  else self.colors.get('unselected'))
        return u''.join((
            self.pos(0, 0), term.normal, attr_l,
            self.left.center(int(math.ceil(float(self.width) / 2))),
            term.normal, attr_r,
            self.right.center(int(math.floor(float(self.width) / 2))),
            term.normal,))

    def move_right(self):
        """ Move selection right, return string suitable for refresh. """
        if self.selection != self.right:
            self.selection = self.right
            return self.refresh()
        return u''

    def move_left(self):
        """ Move selection left, return string suitable for refresh. """
        if self.selection != self.left:
            self.selection = self.left
            return self.refresh()
        return u''

    def toggle(self):
        """ Toggle selection, return string suitable for refresh. """
        if self.selection == self.left:
            self.selection = self.right
        else:
            self.selection = self.left
        return self.refresh()

    @property
    def quit(self):
        """ Whether a 'quit' character has been handled, such as escape. """
        return self._quit



""" Base classes for clients and connections of x/84. """

import array
import errno
import logging
import socket
import threading
import time
import warnings

# local
from x84.bbs.exception import Disconnected
from x84.terminal import spawn_client_session


class BaseClient(object):

    """
    Base class for remote client implementations.

    Instantiated by the corresponding :class:`BaseServer` class.
    """

    #: Override in subclass: a general string identifier for the
    #: connecting protocol (for example, 'telnet', 'ssh', 'rlogin')
    kind = None

    #: maximum unit of data received for each call to socket_recv()
    BLOCKSIZE_RECV = 64

    #: terminal type identifier when not yet negotiated
    TTYPE_UNDETECTED = 'unknown'

    def __init__(self, sock, address_pair, on_naws=None):
        """ Class initializer. """
        self.log = logging.getLogger(self.__class__.__name__)
        self.sock = sock
        self.address_pair = address_pair
        self.on_naws = on_naws
        self.active = True
        self.env = dict([('TERM', self.TTYPE_UNDETECTED),
                         ('LINES', 24),
                         ('COLUMNS', 80),
                         ('connection-type', self.kind),
                         ])
        self.send_buffer = array.array('c')
        self.recv_buffer = array.array('c')
        self.bytes_received = 0
        self.connect_time = time.time()
        self.last_input_time = time.time()

    def close(self):
        """ Close connection with the client. """
        self.shutdown()

    def fileno(self):
        """ File descriptor number of socket. """
        try:
            return self.sock.fileno()
        except socket.error:
            return None

    def input_ready(self):
        """ Whether any data is buffered for reading. """
        return bool(self.recv_buffer.__len__())

    def recv_ready(self):
        """
        Subclass and implement: whether socket_recv() should be called.

        :raises NotImplementedError
        """
        raise NotImplementedError()

    def send(self):
        """
        Send any data buffered and return number of bytes send.

        :raises Disconnected: client has disconnected (cannot write to socket).
        """
        if not self.send_ready():
            warnings.warn('send() called on empty buffer', RuntimeWarning, 2)
            return 0

        ready_bytes = bytes(''.join(self.send_buffer))
        self.send_buffer = array.array('c')

        def _send(send_bytes):
            """
            Inner low-level function for socket send.

            :raises Disconnected: on sock.send error.
            """
            try:
                return self.sock.send(send_bytes)
            except socket.error as err:
                if err.errno in (errno.EDEADLK, errno.EAGAIN):
                    self.log.debug('{self.addrport}: {err} (bandwidth exceed)'
                                   .format(self=self, err=err))
                    return 0
                raise Disconnected('send: {0}'.format(err))

        sent = _send(ready_bytes)
        if sent < len(ready_bytes):
            # re-buffer data that could not be pushed to socket;
            self.send_buffer.fromstring(ready_bytes[sent:])
        return sent

    def send_ready(self):
        """ Whether any data is buffered for delivery. """
        return bool(self.send_buffer.__len__())

    def shutdown(self):
        """
        Shutdown and close socket.

        Called by event loop after client is marked by :meth:`deactivate`.
        """
        try:
            self.sock.shutdown(socket.SHUT_RDWR)
            self.log.debug('{self.addrport}: socket shutdown '
                           '{self.__class__.__name__}'.format(self=self))
        except socket.error:
            pass
        self.active = False
        self.sock.close()

    def socket_recv(self):
        """
        Receive data from socket, returns number of bytes received.

        :raises Disconnect: client has disconnected.
        :rtype: int
        """
        try:
            data = self.sock.recv(self.BLOCKSIZE_RECV)
            recv = len(data)
            if recv == 0:
                raise Disconnected('Closed by client (EOF)')

        except socket.error as err:
            if err.errno == errno.EWOULDBLOCK:
                return 0
            raise Disconnected('socket_recv error: {0}'.format(err))

        self.bytes_received += recv
        self.last_input_time = time.time()
        self.recv_buffer.fromstring(data)
        return recv

    def get_input(self):
        """
        Receive input from client into ``self.recv_buffer``.

        Should be called conditionally when :meth:`input_ready` returns True.
        """
        data = self.recv_buffer.tostring()
        self.recv_buffer = array.array('c')
        return data

    def send_str(self, bstr):
        """ Buffer bytestring for client. """
        self.send_buffer.fromstring(bstr)

    def send_unicode(self, ucs, encoding='utf8'):
        """ Buffer unicode string, encoded for client as 'encoding'. """
        self.send_str(ucs.encode(encoding, 'replace'))

    def is_active(self):
        """ Whether this connection is active (bool). """
        return self.active

    def deactivate(self):
        """ Flag client for disconnection by engine loop. """
        if self.active:
            self.active = False
            self.log.debug('{self.addrport}: deactivated'.format(self=self))

    def idle(self):
        """ Time elapsed since data was last received. """
        return time.time() - self.last_input_time

    def duration(self):
        """ Time elapsed since connection was made. """
        return time.time() - self.connect_time

    @property
    def addrport(self):
        """ IP address and port of connection as string (ip:port). """
        return '%s:%d' % (self.address_pair[0], self.address_pair[1])


class BaseConnect(threading.Thread):

    """ Base class for client connect factories. """

    #: whether this thread is completed. Set to ``True`` to cause an on-connect
    #: thread to forcefully exit.
    stopped = False

    def __init__(self, client):
        """ Class initializer. """
        self.client = client
        threading.Thread.__init__(self)
        self.log = logging.getLogger(self.__class__.__name__)

    def banner(self):
        """ Write data on-connect, callback from :meth:`run`. """
        pass

    def run(self):
        """
        Negotiate a connecting session.

        In the case of telnet and ssh, for example, negotiates and
        inquires about terminal type, telnet options, window size,
        and tcp socket options before spawning a new session.
        """
        try:
            self._set_socket_opts()
            self.banner()
            if self.client.is_active():
                return spawn_client_session(client=self.client)
        except (Disconnected, socket.error) as err:
            self.log.debug('Connection closed: %s', err)
        finally:
            self.stopped = True
        self.client.deactivate()

    def _set_socket_opts(self):
        """
        Set socket non-blocking and enable TCP KeepAlive.

        Callback from :meth:`run`.
        """
        self.client.sock.setblocking(0)
        self.client.sock.setsockopt(
            socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
